"""
Created on 01 March 2016
@author: Mojtaba Haghighatlari
"""
from __future__ import print_function
from builtins import range
import warnings
import os
import time
import pandas as pd
from lxml import objectify, etree

from chemml.utils import std_datetime_str
from chemml.utils import bool_formatter
from chemml.utils import tot_exec_time_str

class Dragon(object):
    """
    An interface to Dragon 6 and 7 chemoinformatics software. Dragon is a commercial software and
    you should provide

    Parameters
    ----------
    version : int, optional (default=7)
        The version of available Dragon on the user's system. (available versions: 6 or 7)

    Weights : list, optional (default=["Mass","VdWVolume","Electronegativity","Polarizability","Ionization","I-State"])
        A list of weights to be used

    blocks : list, optional (default = list(range(1,31)))
        A list of integers as descriptor blocks' id. There are totally 29 and 30 blocks available in
        version 6 and 7, respectively.
        This module is not atimed to cherry pick descriptors in each block.
        For doing so, please use Script Wizard in Drgon GUI.

    external : boolean, optional (default=False)
        If True, include external variables at the end of each saved file.

    Notes
    -----
        The documentation for the rest of parameters can be found in the following links:
            - http://www.talete.mi.it/help/dragon_help/index.html
            - https://chm.kode-solutions.net/products_dragon_tutorial.php
        In the current version, if 3D descriptors are required, we recommend the user to provide .mol2 files that contain
        3D information for the molecules. 
        We also recommend the user to use this class with the following parameters:
        molInput = 'file'
        molfile = path to one SMILES representation file with .smi format or a dictionary of many filepaths
        script = 'new'


    Examples
    --------
        >>> import pandas as pd
        >>> from chemml.chem import Dragon
        >>> drg = Dragon()
        >>> drg.script_wizard(script='new', output_directory='./')
        >>> drg.run()
        >>> df = drg.convert_to_csv(remove=True)
        >>> df = df.drop(['No.','NAME'], axis=1)
    """

    def __init__(self,
                 version=7,
                 molFile='required_required',
                 molInput="file",
                 CheckUpdates=True,
                 SaveLayout=True,
                 PreserveTemporaryProjects=True,
                 ShowWorksheet=False,
                 Decimal_Separator=".",
                 Missing_String="NaN",
                 DefaultMolFormat="1",
                 HelpBrowser="/usr/bin/xdg-open",
                 RejectUnusualValence=False,
                 Add2DHydrogens=False,
                 MaxSRforAllCircuit="19",
                 MaxSR="35",
                 MaxSRDetour="30",
                 MaxAtomWalkPath="2000",
                 LogPathWalk=True,
                 LogEdge=True,
                 Weights=("Mass", "VdWVolume", "Electronegativity", "Polarizability", "Ionization",
                          "I-State"),
                 SaveOnlyData=False,
                 SaveLabelsOnSeparateFile=False,
                 SaveFormatBlock="%b-%n.txt",
                 SaveFormatSubBlock="%b-%s-%n-%m.txt",
                 SaveExcludeMisVal=False,
                 SaveExcludeAllMisVal=False,
                 SaveExcludeConst=False,
                 SaveExcludeNearConst=False,
                 SaveExcludeStdDev=False,
                 SaveStdDevThreshold="0.0001",
                 SaveExcludeCorrelated=False,
                 SaveCorrThreshold="0.95",
                 SaveExclusionOptionsToVariables=False,
                 SaveExcludeMisMolecules=False,
                 SaveExcludeRejectedMolecules=False,
                 blocks=list(range(1, 31)),
                 molInputFormat="SMILES",
                 SaveStdOut=False,
                 SaveProject=False,
                 SaveProjectFile="Dragon_project.drp",
                 SaveFile=True,
                 SaveType="singlefile",
                 SaveFilePath="Dragon_descriptors.txt",
                 logMode="file",
                 logFile="Dragon_log.txt",
                 external=False,
                 fileName=None,
                 delimiter=",",
                 consecutiveDelimiter=False,
                 MissingValue="NaN",
                 RejectDisconnectedStrucuture=False,
                 RetainBiggestFragment=False,
                 DisconnectedCalculationOption="0",
                 RoundCoordinates=True,
                 RoundWeights=True,
                 RoundDescriptorValues=True,
                 knimemode=False):
        if version in (6, 7):
            self.version = version
        else:
            msg = "Only version 6 and 7 are available through this module."
            raise ValueError(msg)
        self.molFile = molFile
        self.molInput = molInput
        self.CheckUpdates = CheckUpdates
        self.PreserveTemporaryProjects = PreserveTemporaryProjects
        self.SaveLayout = SaveLayout
        self.ShowWorksheet = ShowWorksheet
        self.Decimal_Separator = Decimal_Separator
        self.Missing_String = Missing_String
        self.DefaultMolFormat = DefaultMolFormat
        self.HelpBrowser = HelpBrowser
        self.RejectUnusualValence = RejectUnusualValence
        self.Add2DHydrogens = Add2DHydrogens
        self.MaxSRforAllCircuit = MaxSRforAllCircuit
        self.MaxSR = MaxSR
        self.MaxSRDetour = MaxSRDetour
        self.MaxAtomWalkPath = MaxAtomWalkPath
        self.LogPathWalk = LogPathWalk
        self.LogEdge = LogEdge
        self.Weights = Weights
        self.SaveOnlyData = SaveOnlyData
        self.SaveLabelsOnSeparateFile = SaveLabelsOnSeparateFile
        self.SaveFormatBlock = SaveFormatBlock
        self.SaveFormatSubBlock = SaveFormatSubBlock
        self.SaveExcludeMisVal = SaveExcludeMisVal
        self.SaveExcludeAllMisVal = SaveExcludeAllMisVal
        self.SaveExcludeConst = SaveExcludeConst
        self.SaveExcludeNearConst = SaveExcludeNearConst
        self.SaveExcludeStdDev = SaveExcludeStdDev
        self.SaveStdDevThreshold = SaveStdDevThreshold
        self.SaveExcludeCorrelated = SaveExcludeCorrelated
        self.SaveCorrThreshold = SaveCorrThreshold
        self.SaveExclusionOptionsToVariables = SaveExclusionOptionsToVariables
        self.SaveExcludeMisMolecules = SaveExcludeMisMolecules
        self.SaveExcludeRejectedMolecules = SaveExcludeRejectedMolecules
        self.blocks = blocks
        self.molInputFormat = molInputFormat
        self.SaveStdOut = SaveStdOut
        self.SaveProject = SaveProject
        self.SaveProjectFile = SaveProjectFile
        self.SaveFile = SaveFile
        self.SaveType = SaveType
        self.SaveFilePath = SaveFilePath
        self.logMode = logMode
        self.logFile = logFile
        self.external = external
        self.fileName = fileName
        self.delimiter = delimiter
        self.consecutiveDelimiter = consecutiveDelimiter
        self.MissingValue = MissingValue
        self.RejectDisconnectedStrucuture = RejectDisconnectedStrucuture
        self.RetainBiggestFragment = RetainBiggestFragment
        self.DisconnectedCalculationOption = DisconnectedCalculationOption
        self.RoundCoordinates = RoundCoordinates
        self.RoundWeights = RoundWeights
        self.RoundDescriptorValues = RoundDescriptorValues
        self.knimemode = knimemode

    def script_wizard(self, script='new', output_directory='./'):
        """
        The script_wizard is designed to build a Dragon script file. The name and
        the functionality of this function is the same as available Script wizard
        in the Dragon Graphic User Interface.
        Note: All reported nodes are mandatory, except the <EXTERNAL> tag
        Note: Script for version 7 doesn't support fingerprints block

        Parameters
        ----------
        script : string, optional (default="new")
            If "new" start creating a new script from scratch. If you want to load an existing script,
            pass the filename with drs format.

        output_directory : string, optional (default = './')
            the path to the working directory to store output files.

        dragon : xml element
            Dragon script in  xml format.

        drs : string
            Dragon script file name

        data_path : string
            The path+name of saved data file in any format. If saveType is 'block'
            or 'subblock' data_path is just the path to the directory that all data
            files have been saved.
        """
        self.output_directory = output_directory
        if not os.path.exists(self.output_directory):
            os.makedirs(self.output_directory)

        if script == 'new':
            if self.version == 6:
                self.dragon = objectify.Element(
                    "DRAGON",
                    version="%i.0.0" % self.version,
                    script_version="1",
                    generation_date=std_datetime_str('date').replace('-', '/'))

                OPTIONS = objectify.SubElement(self.dragon, "OPTIONS")
                OPTIONS.append(
                    objectify.Element("CheckUpdates", value=bool_formatter(self.CheckUpdates)))
                OPTIONS.append(objectify.Element("SaveLayout", value=bool_formatter(self.SaveLayout)))
                OPTIONS.append(
                    objectify.Element("ShowWorksheet", value=bool_formatter(self.ShowWorksheet)))
                OPTIONS.append(objectify.Element("Decimal_Separator", value=self.Decimal_Separator))
                OPTIONS.append(objectify.Element("Missing_String", value=self.Missing_String))
                OPTIONS.append(objectify.Element("DefaultMolFormat", value=self.DefaultMolFormat))
                OPTIONS.append(objectify.Element("HelpBrowser", value=self.HelpBrowser))
                OPTIONS.append(
                    objectify.Element(
                        "RejectUnusualValence", value=bool_formatter(self.RejectUnusualValence)))
                OPTIONS.append(
                    objectify.Element("Add2DHydrogens", value=bool_formatter(self.Add2DHydrogens)))
                OPTIONS.append(objectify.Element("MaxSRforAllCircuit", value=self.MaxSRforAllCircuit))
                OPTIONS.append(objectify.Element("MaxSR", value=self.MaxSR))
                OPTIONS.append(objectify.Element("MaxSRDetour", value=self.MaxSRDetour))
                OPTIONS.append(objectify.Element("MaxAtomWalkPath", value=self.MaxAtomWalkPath))
                OPTIONS.append(objectify.Element("LogPathWalk", value=bool_formatter(self.LogPathWalk)))
                OPTIONS.append(objectify.Element("LogEdge", value=bool_formatter(self.LogEdge)))
                Weights = objectify.SubElement(OPTIONS, "Weights")
                for weight in self.Weights:
                    if weight not in [
                            "Mass", "VdWVolume", "Electronegativity", "Polarizability", "Ionization",
                            "I-State"
                    ]:
                        msg = "'%s' is not a valid weight type." % weight
                        raise ValueError(msg)
                    Weights.append(objectify.Element('weight', name=weight))
                OPTIONS.append(
                    objectify.Element("SaveOnlyData", value=bool_formatter(self.SaveOnlyData)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveLabelsOnSeparateFile", value=bool_formatter(self.SaveLabelsOnSeparateFile)))
                OPTIONS.append(objectify.Element("SaveFormatBlock", value=self.SaveFormatBlock))
                OPTIONS.append(objectify.Element("SaveFormatSubBlock", value=self.SaveFormatSubBlock))
                OPTIONS.append(
                    objectify.Element("SaveExcludeMisVal", value=bool_formatter(self.SaveExcludeMisVal)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeAllMisVal", value=bool_formatter(self.SaveExcludeAllMisVal)))
                OPTIONS.append(
                    objectify.Element("SaveExcludeConst", value=bool_formatter(self.SaveExcludeConst)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeNearConst", value=bool_formatter(self.SaveExcludeNearConst)))
                OPTIONS.append(
                    objectify.Element("SaveExcludeStdDev", value=bool_formatter(self.SaveExcludeStdDev)))
                OPTIONS.append(objectify.Element("SaveStdDevThreshold", value=self.SaveStdDevThreshold))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeCorrelated", value=bool_formatter(self.SaveExcludeCorrelated)))
                OPTIONS.append(objectify.Element("SaveCorrThreshold", value=self.SaveCorrThreshold))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExclusionOptionsToVariables",
                        value=bool_formatter(self.SaveExclusionOptionsToVariables)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeMisMolecules", value=bool_formatter(self.SaveExcludeMisMolecules)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeRejectedMolecules",
                        value=bool_formatter(self.SaveExcludeRejectedMolecules)))

                DESCRIPTORS = objectify.SubElement(self.dragon, "DESCRIPTORS")
                for i in self.blocks:
                    if i < 1 or i > 29:
                        msg = "block id must be in range 1 to 29."
                        raise ValueError(msg)
                    DESCRIPTORS.append(objectify.Element('block', id="%i" % i, SelectAll="true"))

                MOLFILES = objectify.SubElement(self.dragon, "MOLFILES")
                MOLFILES.append(objectify.Element("molInput", value=self.molInput))
                # if self.molInput == "stdin":
                #     if self.molInputFormat not in ['SYBYL', 'MDL', 'HYPERCHEM', 'SMILES', 'MACROMODEL']:
                #         msg = "'%s' is not a valid molInputFormat. Formats:['SYBYL','MDL','HYPERCHEM','SMILES','MACROMODEL']" % self.molInputFormat
                #         raise ValueError(msg)
                #     MOLFILES.append(objectify.Element("molInputFormat", value=self.molInputFormat))
                if self.molInput == "file":
                    if isinstance(self.molFile, dict):
                        for f in range(1, len(self.molFile) + 1):
                            if os.path.exists(self.molFile[f]['file']):
                                MOLFILES.append(objectify.Element("molFile", value=self.molFile[f]['file']))
                            else:
                                msg = 'file does not exist at '+ self.molFile[f]['file']
                                raise FileNotFoundError(msg)
                    elif isinstance(self.molFile, str):
                        MOLFILES.append(objectify.Element("molFile", value=self.molFile))
                    else:
                        msg = 'Variable molInput can be either a string or a list'
                        raise ValueError(msg)
                else:
                    msg = "The molInput value must be 'file'. 'stdin' is not supported through ChemML"
                    raise ValueError(msg)
                OUTPUT = objectify.SubElement(self.dragon, "OUTPUT")
                OUTPUT.append(objectify.Element("SaveStdOut", value=bool_formatter(self.SaveStdOut)))
                OUTPUT.append(objectify.Element("SaveProject", value=bool_formatter(self.SaveProject)))
                if self.SaveProject:
                    OUTPUT.append(objectify.Element("SaveProjectFile", value=self.SaveProjectFile))
                OUTPUT.append(objectify.Element("SaveFile", value=bool_formatter(self.SaveFile)))
                if self.SaveFile:
                    OUTPUT.append(objectify.Element(
                        "SaveType", value=self.SaveType))  # value = "[singlefile/block/subblock]"
                    OUTPUT.append(
                        objectify.Element(
                            "SaveFilePath", value=self.output_directory + self.SaveFilePath)
                    )  #Specifies the file name for saving results as a plan text file(s), if the "singlefile" option is set; if "block" or "subblock" are set, specifies the path in which results files will be saved.
                OUTPUT.append(objectify.Element("logMode",
                                                value=self.logMode))  # value = [none/stderr/file]
                if self.logMode == "file":
                    OUTPUT.append(
                        objectify.Element("logFile", value=self.output_directory + self.logFile))

                if self.external:
                    EXTERNAL = objectify.SubElement(self.dragon, "EXTERNAL")
                    EXTERNAL.append(objectify.Element("fileName", value=self.fileName))
                    EXTERNAL.append(objectify.Element("delimiter", value=self.delimiter))
                    EXTERNAL.append(
                        objectify.Element(
                            "consecutiveDelimiter", value=bool_formatter(self.consecutiveDelimiter)))
                    EXTERNAL.append(objectify.Element("MissingValue", value=self.MissingValue))
                self._save_script()

            elif self.version == 7:
                self.dragon = objectify.Element(
                    "DRAGON",
                    version="%i.0.0" % self.version,
                    description="Dragon7 - FP1 - MD5270",
                    script_version="1",
                    generation_date=std_datetime_str('date').replace('-', '/'))

                OPTIONS = objectify.SubElement(self.dragon, "OPTIONS")
                OPTIONS.append(
                    objectify.Element("CheckUpdates", value=bool_formatter(self.CheckUpdates)))
                OPTIONS.append(
                    objectify.Element(
                        "PreserveTemporaryProjects",
                        value=bool_formatter(self.PreserveTemporaryProjects)))
                OPTIONS.append(objectify.Element("SaveLayout", value=bool_formatter(self.SaveLayout)))
                #                 OPTIONS.append(objectify.Element("ShowWorksheet", value = bool_formatter(self.ShowWorksheet)))
                OPTIONS.append(objectify.Element("Decimal_Separator", value=self.Decimal_Separator))
                if self.Missing_String == "NaN": self.Missing_String = "na"
                OPTIONS.append(objectify.Element("Missing_String", value=self.Missing_String))
                OPTIONS.append(objectify.Element("DefaultMolFormat", value=self.DefaultMolFormat))
                #                 OPTIONS.append(objectify.Element("HelpBrowser", value = self.HelpBrowser))
                OPTIONS.append(
                    objectify.Element(
                        "RejectDisconnectedStrucuture",
                        value=bool_formatter(self.RejectDisconnectedStrucuture)))
                OPTIONS.append(
                    objectify.Element(
                        "RetainBiggestFragment", value=bool_formatter(self.RetainBiggestFragment)))
                OPTIONS.append(
                    objectify.Element(
                        "RejectUnusualValence", value=bool_formatter(self.RejectUnusualValence)))
                OPTIONS.append(
                    objectify.Element("Add2DHydrogens", value=bool_formatter(self.Add2DHydrogens)))
                OPTIONS.append(
                    objectify.Element(
                        "DisconnectedCalculationOption", value=self.DisconnectedCalculationOption))
                OPTIONS.append(objectify.Element("MaxSRforAllCircuit", value=self.MaxSRforAllCircuit))
                #                 OPTIONS.appendm(objectify.Element("MaxSR", value = self.MaxSR))
                OPTIONS.append(objectify.Element("MaxSRDetour", value=self.MaxSRDetour))
                OPTIONS.append(objectify.Element("MaxAtomWalkPath", value=self.MaxAtomWalkPath))
                OPTIONS.append(
                    objectify.Element("RoundCoordinates", value=bool_formatter(self.RoundCoordinates)))
                OPTIONS.append(
                    objectify.Element("RoundWeights", value=bool_formatter(self.RoundWeights)))
                OPTIONS.append(objectify.Element("LogPathWalk", value=bool_formatter(self.LogPathWalk)))
                OPTIONS.append(objectify.Element("LogEdge", value=bool_formatter(self.LogEdge)))
                Weights = objectify.SubElement(OPTIONS, "Weights")
                for weight in self.Weights:
                    if weight not in [
                            "Mass", "VdWVolume", "Electronegativity", "Polarizability", "Ionization",
                            "I-State"
                    ]:
                        msg = "'%s' is not a valid weight type." % weight
                        raise ValueError(msg)
                    Weights.append(objectify.Element('weight', name=weight))
                OPTIONS.append(
                    objectify.Element("SaveOnlyData", value=bool_formatter(self.SaveOnlyData)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveLabelsOnSeparateFile", value=bool_formatter(self.SaveLabelsOnSeparateFile)))
                OPTIONS.append(objectify.Element("SaveFormatBlock", value=self.SaveFormatBlock))
                OPTIONS.append(objectify.Element("SaveFormatSubBlock", value=self.SaveFormatSubBlock))
                OPTIONS.append(
                    objectify.Element("SaveExcludeMisVal", value=bool_formatter(self.SaveExcludeMisVal)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeAllMisVal", value=bool_formatter(self.SaveExcludeAllMisVal)))
                OPTIONS.append(
                    objectify.Element("SaveExcludeConst", value=bool_formatter(self.SaveExcludeConst)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeNearConst", value=bool_formatter(self.SaveExcludeNearConst)))
                OPTIONS.append(
                    objectify.Element("SaveExcludeStdDev", value=bool_formatter(self.SaveExcludeStdDev)))
                OPTIONS.append(objectify.Element("SaveStdDevThreshold", value=self.SaveStdDevThreshold))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeCorrelated", value=bool_formatter(self.SaveExcludeCorrelated)))
                OPTIONS.append(objectify.Element("SaveCorrThreshold", value=self.SaveCorrThreshold))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExclusionOptionsToVariables",
                        value=bool_formatter(self.SaveExclusionOptionsToVariables)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeMisMolecules", value=bool_formatter(self.SaveExcludeMisMolecules)))
                OPTIONS.append(
                    objectify.Element(
                        "SaveExcludeRejectedMolecules",
                        value=bool_formatter(self.SaveExcludeRejectedMolecules)))
                OPTIONS.append(
                    objectify.Element(
                        "RoundDescriptorValues", value=bool_formatter(self.RoundDescriptorValues)))

                DESCRIPTORS = objectify.SubElement(self.dragon, "DESCRIPTORS")
                for i in self.blocks:
                    if i < 1 or i > 30:
                        msg = "block id must be in range 1 to 30."
                        raise ValueError(msg)
                    DESCRIPTORS.append(objectify.Element('block', id="%i" % i, SelectAll="true"))

                MOLFILES = objectify.SubElement(self.dragon, "MOLFILES")
                MOLFILES.append(objectify.Element("molInput", value=self.molInput))
                # if self.molInput == "stdin":
                #     if self.molInputFormat not in [
                #             'SYBYL', 'MDL', 'HYPERCHEM', 'SMILES', 'CML', 'MACROMODEL'
                #     ]:
                #         msg = "'%s' is not a valid molInputFormat. Formats:['SYBYL','MDL','HYPERCHEM','SMILES','CML','MACROMODEL']" % self.molInputFormat
                #         raise ValueError(msg)
                #     MOLFILES.append(objectify.Element("molInputFormat", value=self.molInputFormat))
                if self.molInput == "file":
                    if isinstance(self.molFile, dict):
                        for f in range(1, len(self.molFile) + 1):
                            if os.path.exists(self.molFile[f]['file']):
                                MOLFILES.append(objectify.Element("molFile", value=self.molFile[f]['file']))
                            else:
                                msg = "file not found at "+ self.molFile[f]['file']
                                raise FileNotFoundError(msg)
                    elif isinstance(self.molFile, str):
                        MOLFILES.append(objectify.Element("molFile", value=self.molFile))
                    else:
                        msg = 'Variable molFile can be either a string or a list'
                        raise ValueError(msg)
                else:
                    msg = "The molInput value must be 'file'. 'stdin' is not supported through ChemML"
                    raise ValueError(msg)
                OUTPUT = objectify.SubElement(self.dragon, "OUTPUT")
                OUTPUT.append(objectify.Element("knimemode", value=bool_formatter(self.knimemode)))
                OUTPUT.append(objectify.Element("SaveStdOut", value=bool_formatter(self.SaveStdOut)))
                OUTPUT.append(objectify.Element("SaveProject", value=bool_formatter(self.SaveProject)))
                if self.SaveProject:
                    OUTPUT.append(objectify.Element("SaveProjectFile", value=self.SaveProjectFile))
                OUTPUT.append(objectify.Element("SaveFile", value=bool_formatter(self.SaveFile)))
                if self.SaveFile:
                    OUTPUT.append(objectify.Element(
                        "SaveType", value=self.SaveType))  # value = "[singlefile/block/subblock]"
                    OUTPUT.append(
                        objectify.Element(
                            "SaveFilePath", value=self.output_directory + self.SaveFilePath)
                    )  #Specifies the file name for saving results as a plan text file(s), if the "singlefile" option is set; if "block" or "subblock" are set, specifies the path in which results files will be saved.
                OUTPUT.append(objectify.Element("logMode",
                                                value=self.logMode))  # value = [none/stderr/file]
                if self.logMode == "file":
                    OUTPUT.append(
                        objectify.Element("logFile", value=self.output_directory + self.logFile))

                if self.external:
                    EXTERNAL = objectify.SubElement(self.dragon, "EXTERNAL")
                    EXTERNAL.append(objectify.Element("fileName", value=self.fileName))
                    EXTERNAL.append(objectify.Element("delimiter", value=self.delimiter))
                    EXTERNAL.append(
                        objectify.Element(
                            "consecutiveDelimiter", value=bool_formatter(self.consecutiveDelimiter)))
                    EXTERNAL.append(objectify.Element("MissingValue", value=self.MissingValue))
                self._save_script()
        else:
            doc = etree.parse(script)
            self.dragon = etree.tostring(doc)  # dragon script : dragon
            self.dragon = objectify.fromstring(self.dragon)
            objectify.deannotate(self.dragon)
            etree.cleanup_namespaces(self.dragon)
            if self.dragon.attrib['version'][0] not in ['6', '7']:
                msg = "Dragon script is not labeled to the newest vesions of Dragon, 6 or 7. This may causes some problems."
                warnings.warn(msg, RuntimeWarning)
            mandatory_nodes = ['OPTIONS', 'DESCRIPTORS', 'MOLFILES', 'OUTPUT']
            reported_nodes = [element.tag for element in self.dragon.iterchildren()]
            if not set(reported_nodes).issuperset(set(mandatory_nodes)):
                msg = "Dragon script doesn't contain all required nodes, which are:%s" % str(
                    mandatory_nodes)
                raise ValueError(msg)
            self.drs = script
        self.data_path = self.dragon.OUTPUT.SaveFilePath.attrib['value']
        return self

    def _save_script(self):
        # objectify.deannotate(self.dragon)
        etree.cleanup_namespaces(self.dragon)
        self.drs_name = 'Dragon_script.drs'
        with open(os.path.join(self.output_directory, self.drs_name), 'w') as outfile:
            outfile.write(etree.tostring(self.dragon, pretty_print=True).decode())

    def printout(self):
        objectify.deannotate(self.dragon)
        etree.cleanup_namespaces(self.dragon)
        print(objectify.dump(self.dragon))

    def run(self):
        t0 = time.time()
        print("running Dragon%i ..." % self.version)
        os_ret = os.system('nohup dragon%sshell -s %s' %
                           (self.version, os.path.join(self.output_directory, self.drs_name)))
        if os_ret != 0:
            msg = "Oops, dragon%ishell command didn't work! Are you sure Dragon%i software is installed on your machine?" % (
                self.version, self.version)
            raise ImportError(msg)

        # execution time
        tmp_str = tot_exec_time_str(t0)
        print("... Dragon job completed in %s"%tmp_str)

        # print subprocess.check_output(['nohup dragon%sshell -s %s'%(self.version,self.drs)])

    def convert_to_csv(self, remove=True):
        """
        This function converts the tab-delimited txt file from Dragon to pandas dataframe.
        Note that this process might require large memory based on the number of data points and features.

        Parameters
        ----------
        remove : bool, optional (default = True)
            if True, the original descriptors file (Dragon_descriptors.txt) will be removed.

        Returns
        -------
        features : pandas DataFrame
            The 2D dataframe of the descriptors. Note that the first two columns are 'No.' and 'NAME'.

        """
        # convert to csv file
        t0 = time.time()
        print("converting output file to csv format ...")
        df = pd.read_csv(self.data_path, sep=None, engine='python')
        # df = df.drop(['No.', 'NAME'], axis=1)

        # execution time
        tmp_str = tot_exec_time_str(t0)

        # remove original tab delimited file
        if remove:
            os.remove(self.data_path)
            self.data_path = None

        print("... conversion completed in %s"%tmp_str)

        return df





"""
This module provides interactive implementation of active learning algorithms to query optimal number of data points.
"""

from __future__ import print_function
from builtins import range

import warnings
import types
import copy

import numpy as np
import pandas as pd

from sklearn.model_selection import ShuffleSplit, KFold
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA


class ActiveLearning(object):
    """
    The implementation of active learning of regression models using BEMCM and QBC methods and approaches for distribution shift alleviations.
    This algorithm assumes that you have a pool of unlabeled data points and a limited budget to label them.
    Thus, we combine the efficiency of the machine learning models with our active learning approach to suggest
    optimal number of calculations to provide labeled data.

    The implementation of this algorithm follows an interactive approach.
    In other words, we often ask you to provide labels for the selected data points.

    Parameters
    ----------
    model_creator : FunctionType
        It's a function that returns the model. We call this function a couple of times during the search to build
        fresh models with random weights.
        Note that you should also compile your model inside the function. We don't provide options to compile the model.
        The compile (e.g., for Keras models) defines the loss function, the optimizer/learning rate, and the metrics.

    U : array-like
        The features/descriptors of unlabeled candidates that are available to be labeled.

    target_layer : str or list or FunctionType
        If str, it's the name of a layer of the Keras model that is linearly mapped to the outputs.
        If list, it's a list of str that each element corresponds to the name of layers.
        If a function, it should be able to receive a model that will be created using the 'model_creator' and the X inputs,
        and returns the outputs of the linear layer.

    train_size : int, optional (default = 100)
        It represents the absolute number of train samples that must be selected as the initial training set.
        The search will begin with this many training samples and labels are required immediately.
        Please choose a number based on:
            - your budget.
            - the minumum number that you think is enough to train your model.

    test_size : int, optional (default = 100)
        It represents the absolute number of test samples that will be held out for the evaluation of the model in all
        rounds of your active learning search.
        Note the test set will be acquired before search begins and won't be updated later during search.
        Please choose a number based on:
            - your budget.
            - the diversity of the pool of candidates (U).

    test_type : str, optional (default = 'passive')
        The value must be either 'passive' or 'active'.
        If passive, test set will be sampled randomly at the initialization.
        If active, test set will be sampled randomly at each round.

    batch_size : list, optional (default = [10])
        This is a list of maxumum three non-negative int values. Each value specifies the number of data points that
        our active learning approaches should query every round. The order of active learning approaches are as follows:
            - Batch Expected Model Change Maximization (BEMCM)
            - Query By Committee (QBC)
            - Distribution Shift Alleviation (DSA)

        Note that the last method (i.e., DSA) is a complement to the first two methods and can not be specified alone.

    history : int, optional (default = 2)
        This parameter must be an integer and greater than one. It specifies the number of previous active learning
        rounds to memorize for the distribution shift alleviation (DSA) approach.

    Attributes
    ----------
    queries : list
        This list provides information regarding the indices of queried candidates.
        for each element of the list:
            - The index-0 is a short description.
            - The index-1 is an array of indices.

    query_number : int
        The number of rounds you have run the active learning search method.

    U_indices : ndarray
        This is an array of the remaining unlabeled indices.

    train_indices : ndarray
        This is an array of all candidates' indices that are used as the training data.

    test_indices : ndarray
        This is an array of all candidates' indices that are used as the test data.

    Y_pred : ndarray
        The predicted Y values at the current stage. These values will be updated after each run of `search` method.

    results : pandas.DataFrame
        The final results of the active learning approach.

    random_results : pandas.DataFrame
        The final results of the random search.


    Methods
    -------
    initialize
    deposit
    search
    random_search
    visualize
    get_target_layer

    Notes
    -----
    - You won't be able to resume the search unless you deposit the requested labeled data.

    """

    def __init__(self, model_creator, U, target_layer, train_size=100, test_size=100,
                 test_type='passive', batch_size=[10], history=2):
        self.model_creator = model_creator
        self.U = U
        self.target_layer = target_layer
        self.train_size = train_size
        self.test_size = test_size
        self.test_type = test_type
        self.batch_size = batch_size
        self.history = history
        self._fit()

    def _X_train(self):
        """ We don't want to keep a potentially big matrix in the memory."""
        return self.U[self.train_indices]

    def _X_test(self):
        """ We don't want to keep a potentially big matrix in the memory."""
        return self.U[self.test_indices]

    @property
    def queries(self):
        return self._queries

    @property
    def X_train(self):
        return self._X_train()

    @property
    def X_test(self):
        return self._X_test()

    @property
    def Y_train(self):
        return self._Y_train

    @property
    def Y_test(self):
        return self._Y_test

    @property
    def Y_pred(self):
        return self._Y_pred

    @property
    def results(self):
        results_header = ['num_query', 'num_training', 'num_test',
                               'mae', 'mae_std', 'rmse', 'rmse_std', 'r2', 'r2_std']
        df = pd.DataFrame(self._results, columns=results_header)
        return df

    @property
    def random_results(self):
        results_header = ['num_query', 'num_training', 'num_test',
                               'mae', 'mae_std', 'rmse', 'rmse_std', 'r2', 'r2_std']
        df = pd.DataFrame(self._random_results, columns=results_header)
        return df

    def _fit(self):
        # np array the input U
        self.U = np.array(self.U)
        self.U_size = self.U.shape[0]

        # Todo: support for sklearn linear models
        if not isinstance(self.model_creator, types.FunctionType):
            msg = "The parameter 'model' only accepts a keras model in the current version."
            raise TypeError(msg)

        # check int types
        if not isinstance(self.U_size, int) or not isinstance(self.train_size, int) or \
            not isinstance(self.test_size, int):
            msg = "The parameters 'train_size' and 'test_size'must be int."
            raise TypeError(msg)

        # check batch size
        flag = False
        if isinstance(self.batch_size, list):
            if len(self.batch_size) <= 3:
                b = np.array([type(i)==int for i in self.batch_size])   # all int
                p = np.array([i>=0 for i in self.batch_size])   # all non-negative
                g = np.array([i>0 for i in self.batch_size])   # any greater than zero
                if b.all() and p.all() and g.any():
                    flag = True

        if not flag:
            msg = "The parameter'batch_size' must be a list of maximum 3 non-negative int values."
            raise TypeError(msg)
        elif len(self.batch_size)==3 and self.batch_size[2]>0 and sum(self.batch_size[:2])==0:
            msg = "The third value can not be the only positive value and it always accompany the first two methods."
            raise ValueError(msg)

        # check history
        flag = False
        if isinstance(self.history, int):
            if self.history > 1:
                flag = True
                self._history = np.zeros(self.U_size*self.history).reshape(self.U_size, self.history)

        if not flag:
            msg = "The parameter `history` must be a positive int and greater than 1."
            raise ValueError(msg)

        # check the number of train and test sets
        if self.train_size >= self.U_size or self.test_size >= self.U_size or \
                (self.train_size+self.test_size) >= self.U_size:
            msg = "The train and test size and their sum must be less than the number of unlabeled data (U)"
            raise ValueError(msg)

        # check the test_type value
        if self.test_type not in ('active', 'passive'):
            msg = "The parameter 'test_type' must be either 'active' or 'passive'."
            raise ValueError(msg)


        # other attributes
        self._queries = []
        self.qbc_queries = []
        self.dsa_queries = []
        self.bemcm_queries = []
        # all indices are numpy arrays
        self.train_indices = np.array([])
        self.test_indices = np.array([])
        self.initial_test_indices = np.array([])
        self.U_indices = np.array(range(self.U_size))
        self.query_number = 0
        self._Y_train = None
        self._Y_test = None
        self._Y_pred = None
        self._results = []
        self._random_results = []
        self.attributes = ['queries','train_indices','test_indices',
                           'query_number',
                           'X_train','X_test','Y_train','Y_test', 'Y_pred',
                           'results', 'random_results']

        self.lr = 0

    def get_target_layer(self, model, X):
        """
        The main function to get the latent features from the linear layer of the keras model.

        Returns
        -------
        target_layer : array-like
            The concatenated array of the specified hidden layers by parameter `target_layer`.
        """
        # import keras here
        from tensorflow.keras import backend as K

        # inputs
        inp = model.input
        if isinstance(inp, list):
            if not isinstance(X, list):
                msg = "The input must be a list of arrays."
                raise ValueError(msg)
        else:
            if isinstance(X, list):
                msg = "Only one input array is required."
                raise ValueError(msg)
            else:
                # list of inp is required for K.function mapping
                inp = [inp]
                # if input is not a list should become a list
                X = [X]

        # outputs
        if isinstance(self.target_layer, str):
            out = [model.get_layer(self.target_layer).output]
        elif isinstance(self.target_layer, list):
            out = [model.get_layer(name).output for name in self.target_layer]
        else:
            msg = "The parameter 'linear_layer' must be str, list of str or a function."
            raise ValueError(msg)

        # define mapping function
        g = K.function(inp, out)

        # find and concatenate target layers
        target_layers = g(X)
        target_layers = np.concatenate(target_layers, axis=-1)

        return target_layers

    def initialize(self,random_state=90):
        """
        The function to initialize the training and test set for the search.
        You can run this function only once before starting the search.

        Parameters
        ----------
        random_state : int or RandomState, optional (default = 90)
            The random state will be directly passed to the sklearn.model_selection.ShuffleSplit
            extra info at: https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.ShuffleSplit.html#sklearn-model-selection-shufflesplit

        Returns
        -------
        train_indices : array-like
            The training set indices (Python 0-index) from the pool of candidates (U).
            This is a 1D array.

        test_indices : array-like
            The test set indices (Python 0-index) from the pool of candidates (U).

        """
        if len(self._queries) != 0:
            train_indices = []
            test_indices = []
            for q_ind in self._queries:
                if q_ind[0] == 'test set':
                    test_indices = q_ind[1]
                else:
                    train_indices = q_ind[1]
            if len(train_indices)>0 or len(test_indices)>0 :
                msg = "The class has been already initialized! You still have to deposit %i more training data and " \
                      "%i more test data. The same indices are returned respectively." % (len(train_indices), len(test_indices))
                warnings.warn(msg)
                return train_indices, test_indices
        elif len(self.train_indices) != 0:
            msg = "The class has been already initialized and it can not be initialized again!"
            raise ValueError(msg)

        ss = ShuffleSplit(n_splits=1, test_size=self.test_size, train_size=self.train_size, random_state=random_state)
        for train_indices, test_indices in ss.split(range(len(self.U_indices))):
            self._queries.append(['initial training set', train_indices])
            self._queries.append(['test set', test_indices])
            return train_indices, test_indices

    def ignore(self, indices):
        """
        If you found out that the experimental setup or computational research on some of the candidates
        is not feasible, just pass a list of their indices here and we remove them from the list of queries.

        Parameters
        ----------
        indices : array-like
            A 1D array of all the indices that should be removed from the list of `queries`.

        """
        indices = np.array(indices)
        # check the dimension of ind be one
        if indices.ndim != 1 :
            msg = "The dimension of 'indices' must be one, i.e., a list, tuple, or array of indices is requested."
            raise ValueError(msg)

        if len(self._queries) > 0:
            for query in self._queries:
                query[1] = np.array([i for i in query[1] if i not in indices])
        else:
            msg = "The list of queries is already empty."
            raise ValueError(msg)

    def deposit(self, indices, Y):
        """
        This function helps you to deposit the data for candidates that were queried by initialize or search functions.

        Parameters
        ----------
        indices: array-like
            A 1-dimensional array of indices that was queried by initialize or search methods.
            You can deposit the data partially and it doesn't have to be the entire array that is queried.

        Y: array-like
            The 2-dimensional labels of the data points as it will be used for the training of the model.
            The first dimension of the array should be equal to the number of indices.
            Y must be at least 2 dimensional.

        Returns
        -------

        check : bool
            True, if deposited properly. False, otherwise.

        """
        indices = np.array(indices)
        Y = np.array(Y)
        # check if Y is 2 dimensional
        if Y.ndim != 2:
            msg = "The labels, 'Y', must be 2 dimensional."
            raise ValueError(msg)

        # check if the length of indices and labels are same
        if indices.shape[0] != Y.shape[0]:
            msg = "The first dimension of the input Y should be equal to the number of indices."
            raise ValueError(msg)

        # check the dimension of ind be one
        if indices.ndim != 1 :
            msg = "The dimension of 'indices' must be one: array, list, or tuple of indices would suffice."
            raise ValueError(msg)

        # check if indices are unique
        if len(set(indices)) != len(indices):
            msg = "The indices are not unique. This causes duplicate data and thus results in biased training."
            raise ValueError(msg)

        if len(self._queries) > 0:
            match_flag = False
            self.last_deposited_indices_ = np.array([])
            for query in self._queries:
                ind_in_q = np.where(np.in1d(indices, query[1], assume_unique=True))[0] # the position of matched numbers in the indices
                if len(ind_in_q)>0:
                    match_flag = True
                    if query[0] == 'test set':      # the only case that we query test data
                        if self._Y_test is None:
                            self._Y_test = Y[ind_in_q]
                        else:
                            self._Y_test = np.append(self._Y_test, Y[ind_in_q], axis=0)
                        # update test_indices and U_indices
                        settled_inds = np.array(indices[ind_in_q])
                        self.last_deposited_indices_ = np.append(self.last_deposited_indices_, settled_inds).astype(int)
                        self.test_indices = np.append(self.test_indices, settled_inds).astype(int)           # array of all indices
                        self.initial_test_indices = np.append(self.initial_test_indices, settled_inds).astype(int)
                        self.U_indices = np.array([i for i in self.U_indices if i not in settled_inds])
                    else:
                        if self._Y_train is None:
                            self._Y_train = Y[ind_in_q]
                        else:
                            self._Y_train = np.append(self._Y_train, Y[ind_in_q], axis=0)
                        # update train_indices and U_indices
                        settled_inds = np.array(indices[ind_in_q])
                        self.last_deposited_indices_ = np.append(self.last_deposited_indices_, settled_inds).astype(int)
                        self.train_indices = np.append(self.train_indices, settled_inds).astype(int)
                        self.U_indices = np.array([i for i in self.U_indices if i not in settled_inds])
                    # update q_ind and thus _queries
                    query[1] = np.array([i for i in query[1] if i not in settled_inds])
            if not match_flag:
                msg = "Can't match the indices with queries."
                raise ValueError(msg)
            else:
                msg = "we stored %i of passed indices. A list of them is in the 'last_deposited_indices_' attribute."%len(self.last_deposited_indices_)
                print(msg)
                # update queries
                self._update_queries()
                _ = self._update_train_test()
                return True
        else:
            msg = "The `queries` is empty. Can't deposit data if it's not been queried."
            warnings.warn(msg)
            return False

    def _update_queries(self):
        """
        This function removes empty lists of queries.

        """
        self._queries = [q for q in self._queries if len(q[1])>0]

    def _update_train_test(self):
        """
        This function take care of the test_type parameter.

        """
        if self.test_type == 'passive':
            return True
        if len(self._queries) > 0:
            return True
        else:
            # active test split
            all_indices = np.concatenate([self.train_indices, self.test_indices], axis=0)
            all_y = np.concatenate([self._Y_train, self._Y_test], axis=0)
            # select randomly
            ss = ShuffleSplit(n_splits=1, test_size=self.test_size, train_size=None, random_state=90)
            for train_indices, test_indices in ss.split(all_indices):
                # test
                self._Y_test = all_y[test_indices]
                self.test_indices = all_indices[test_indices]
                # train
                self._Y_train = all_y[train_indices]
                self.train_indices = all_indices[train_indices]

    def _scaler(self, scale):
        """
        The internal function to manage the X and Y scalers.

        Returns
        -------
        scaler_x : scaler
            The X scaler.

        scaler_y : scaler
            The Y scaler.

        """
        if isinstance(scale, list):
            if len(scale) == 2:
                Xscaler = scale[0]
                Yscaler = scale[1]
                return Xscaler, Yscaler
            else:
                msg = "The length of the parameter 'scale' must be two. The first scaler is for the X array and the second one for Y."
                raise ValueError(msg)
        else:
            if scale:
                return StandardScaler(), StandardScaler()
            else:
                return None, None

    def search(self, n_evaluation=3, ensemble='bootstrap', n_ensemble=4, normalize_input=True, normalize_internal=False, random_state=90, **kwargs):
        """
        The main function to start or continue an active learning search.
        The bootstrap approach is used to generate an ensemble of models that estimate the prediction
        distribution of the candidates' labels.

        Parameters
        ----------
        n_evaluation : int, optional (default = 3)
            number of times to repeat training of the model and evaluation on test set.

        ensemble : str, optional (default = 'bootstrap')
            The sampling method to create n ensembles and estimate the predictive distributions.
                - 'bootstrap': standard bootstrap method (random choice with replacement)
                - 'shuffle' : sklearn.model_selection.ShuffleSplit
                - 'kfold' : sklearn.model_selection.KFold
            The 'shuffle' and 'kfold' methods draw samples that are smaller than training set.

        n_ensemble : int, optional (default = 5)
            The size of the ensemble based on bootstrapping approach.

        normalize_input : bool or list, optional (default = True)
            if True, sklearn.preprocessing.StandardScaler will be used to normalize X and Y before training.
            You can also pass a list of two scaler instances that perform sklearn-style fit_transform and transform methods
            for the X and Y, respectively.

        normalize_internal : bool, optional (default = False)
            if True, the internal variables for estimation of gradients will be normalized.

        random_state : int or RandomState, optional (default = 90)
            The random state will be directly passed to the sklearn.model_selection.KFold or ShuffleSplit
            Additional info at: https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.KFold.html

        kwargs
            Any argument (except input data) that should be passed to the model's fit method.

        Returns
        -------
        train_indices : array-like
            The training set indices (Python 0-index) from the pool of candidates (U).
            This is a 1D array.

        """
        # check if queries are provided
        if len(self._queries) > 0 :
            msg = "The requested data must be provided first. Check the 'queries' attribute for the info regarding the indices of the queried candidates."
            raise ValueError(msg)

        # assign batch size (length of batch size is not always 3, otherwise we could enumerate the list directly)
        bemcm = 0
        qbc = 0
        dsa = 0
        for i, s in enumerate(self.batch_size):
            if i==0:
                bemcm = s
            elif i==1:
                qbc = s
            elif i==2:
                dsa = s

        # get input data
        X_tr = self._X_train()
        X_te = self._X_test()
        Y_tr = copy.deepcopy(self._Y_train)
        Y_te = copy.deepcopy(self._Y_test)

        # scale
        X_scaler, Y_scaler = self._scaler(normalize_input)
        if X_scaler is not None:
            # scale X arrays
            Utr = X_scaler.fit_transform(self.U)
            X_tr = X_scaler.transform(X_tr)
            X_te = X_scaler.transform(X_te)
            # scale Y
            Y_tr = Y_scaler.fit_transform(Y_tr)
        else:
            Utr = self.U

        # assert not (Utr == self.U).all()  # run just for test

        # make sure the data is not overwritten
        # assert not (X_tr == self.U[self.train_indices]).all()   # run just for test
        assert not (Y_tr == self._Y_train).all()

        # training and evaluation
        it_results = {'mae':[], 'rmse':[], 'r2':[]}
        Y_U_pred_df = pd.DataFrame()  # empty dataframe to collect f(U) at each iteration
        learning_rate = []
        lin_layers = {}
        for it in range(n_evaluation):
            model = self.model_creator()
            model, _, mae, rmse, r2 = self._train_predict_evaluate(model,
                                                                           [X_tr, Y_tr, X_te],
                                                                           Y_scaler,
                                                                           Y_te,
                                                                           **kwargs)
            # Todo: how can we support multioutput?
            # predict Y of remaining U, f(Utr)
            if Y_scaler is not None:
                Y_U_pred_df[it] = Y_scaler.inverse_transform(model.predict(Utr)).reshape(-1,)

            # calculate the linear layer, phi(U), and collect lr for bemcm approach
            if bemcm:
                lin_layers[it] = self.get_target_layer(model, Utr[self.U_indices])
                # if it == 0:
                #     lin_layer = self.get_target_layer(model, Utr[self.U_indices])
                # else:
                #     temp = self.get_target_layer(model, Utr[self.U_indices])
                #     lin_layer = lin_layer + temp

                assert lin_layers[it].shape[0] == self.U_indices.shape[0]

                # collect lr
                from tensorflow.keras import backend as K
                learning_rate.append(K.eval(model.optimizer.lr))

            # metrics
            it_results['mae'].append(mae)
            it_results['rmse'].append(rmse)
            it_results['r2'].append(r2)
            del model

        # store evaluation results
        results_temp = [self.query_number, len(self.train_indices), len(self.test_indices)]
        for metric in ['mae', 'rmse', 'r2']:
            results_temp.append(np.mean(np.array(it_results[metric]), axis=0))
            results_temp.append(np.std(np.array(it_results[metric]), axis = 0, ddof=0))
        self._results.append(results_temp)

        # find Y predictions of all candidates U
        assert Y_U_pred_df.shape == (self.U_size, n_evaluation)
        self._Y_pred = Y_U_pred_df.mean(axis=1).values.reshape(-1, 1)
        # transform back to scaled
        if Y_scaler is not None:
            fU_preds_scaled = Y_scaler.transform(self._Y_pred)
        else:
            fU_preds_scaled = self._Y_pred
        assert self._Y_pred.shape == (self.U_size, 1)
        del Y_U_pred_df

        # find linear layer input and learning rate for bemcm approach
        if bemcm:
            best_ind = np.argmin(it_results['mae'])     # the order of lin_layer might be different from one model to another model
            lin_layer = lin_layers[best_ind]
            # lin_layer = lin_layer/float(n_evaluation)   # shape: (m,d)

            # scale linear layer
            if normalize_internal:
                scaler = StandardScaler()
                lin_layer = scaler.fit_transform(lin_layer)

            # avg of learning rates
            alpha = float(np.mean(learning_rate))
            self.lr = alpha

        # Ensemble
        if ensemble=='kfold' and n_ensemble>1:
            cv = KFold(n_splits=n_ensemble, shuffle=True, random_state=random_state)
            g = cv.split(X_tr)
        elif ensemble=='shuffle':
            cv = ShuffleSplit(n_splits=n_ensemble, train_size = X_tr.shape[0]-1, test_size= None, random_state=random_state)
            g = cv.split(X_tr)
        elif ensemble == 'bootstrap' or n_ensemble == 1:
            g = None
        else:
            msg = "You must select between 'bootstrap', 'kfold' or 'shuffle' sampling methods with the `n_ensemble` greater than zero."
            raise ValueError(msg)

        for it in range(n_ensemble):
            if g is None:
                train_index = np.random.choice(range(len(X_tr)), size=len(X_tr), replace=True)
            else:
                train_index, _ = next(g)
            Xtr = X_tr[train_index]
            Ytr = Y_tr[train_index]
            # model
            model = self.model_creator()
            model, Z_U_pred, _, _, _ = self._train_predict_evaluate(model,
                                                                    [Xtr,Ytr,Utr],
                                                                    None,   # don't inverse_transform preds
                                                                    False,
                                                                    **kwargs)
            deviation = fU_preds_scaled - Z_U_pred     # shape: (m,1)

            # collect the bootstrap deviation from actual predictions
            if it==0:
                deviations = deviation
            else:
                deviations = np.append(deviations, deviation, axis=1)

            del Xtr, Z_U_pred, model
        del X_tr, Y_tr, Utr      # from now on we only need deviations and lin_layer

        assert deviations.shape == (self.U_size, n_ensemble)

        # scale deviations
        if normalize_internal:
            scaler = StandardScaler()
            deviations = scaler.fit_transform(deviations)

        i_dsa_queries = []
        if dsa:
            i_dsa_queries = self._dsa_y_dist()
            # i_dsa_queries = self._dsa_test_y()
            # i_dsa_queries = self._dsa_unlabeled(deviations)

        i_qbc_queries = []
        if qbc and not bemcm:     # bemcm can cover for duplicates in all of the approaches
            i_qbc_queries = self._qbc(deviations[self.U_indices],i_dsa_queries)
        elif qbc:
            i_qbc_queries = self._qbc(deviations[self.U_indices],None)

        i_bemcm_queries = []
        if bemcm:
            i_bemcm_queries = self._bemcm(deviations[self.U_indices], lin_layer, alpha, normalize_internal,
                                              i_qbc_queries+i_dsa_queries)

        i_queries = np.unique(i_qbc_queries+i_dsa_queries+i_bemcm_queries)
        assert len(i_queries) == sum(self.batch_size)

        # increase global query number by one
        self.query_number += 1
        if qbc:
            self.qbc_queries += [self.U_indices[i] for i in i_qbc_queries]
        if dsa:
            self.dsa_queries += [self.U_indices[i] for i in i_dsa_queries if i not in i_qbc_queries]
        if bemcm:
            self.bemcm_queries += [self.U_indices[i] for i in i_bemcm_queries]

        # find original indices and update queries
        _queries = np.array([self.U_indices[i] for i in i_queries])

        self._queries.append(['batch #%i'%self.query_number, _queries])
        return _queries

    def _qbc(self, deviations, former_queries):
        """
        qbc approach
        """
        votes = pd.DataFrame(deviations)
        votes['sigma'] = deviations.std(axis=1)
        votes['ind'] = votes.index
        votes.sort_values('sigma', ascending=False, inplace=True)

        # check if qbc should compensate underselection by dsa approach
        if former_queries is None:
            qbc_size = self.batch_size[1]
            former_queries = []
        else:
            former_queries = np.unique(former_queries)
            qbc_size = sum(self.batch_size) - len(former_queries)

        # start query
        i_qbc_queries = []
        n = 1
        while len(i_qbc_queries) < qbc_size:
            select = list(votes.head(n)['ind'])
            while select[-1] in former_queries:
                n += 1
                select = list(votes.head(n)['ind'])
            i_qbc_queries += select[-1:]
            n += 1
        return i_qbc_queries

    def _dsa_y_dist(self):
        """
        dsa approach based on the distribution of the test data and predicted y values
        """
        # test distribuiton
        f = pd.DataFrame(self._Y_test, columns=['yt'])
        f['ind'] = f.index
        out, bins = pd.cut(f.yt, 100, retbins=True)
        groups = f.groupby(['ind', out])
        _ytest_dist = pd.DataFrame(groups.size().unstack().sum())
        _ytest_dist.sort_values(0, ascending=False, inplace=True)
        _ytest_dist['prob'] = _ytest_dist[0]/sum(_ytest_dist[0])

        # train distribution
        f = pd.DataFrame(self._Y_train, columns=['yt'])
        f['ind'] = f.index
        groups = f.groupby(['ind', pd.cut(f.yt, bins)])
        _ytrain_dist = pd.DataFrame(groups.size().unstack().sum())
        _ytrain_dist.sort_values(0, ascending=False, inplace=True)
        _ytrain_dist['prob'] = _ytrain_dist[0]/sum(_ytrain_dist[0])
        del f

        # difference in distribution
        self._dist_shift = pd.DataFrame(_ytest_dist['prob'] - _ytrain_dist['prob'])
        self._dist_shift.sort_values('prob', ascending=False, inplace=True)

        # find n selection per bin for underrepresented points in training data
        intervals = tuple(self._dist_shift[self._dist_shift['prob'] > 0].index)
        if len(intervals) == 0:
            return []   # if no distribution shift was diagnosed this method will be ineffective
        n_selection_p_bin = int(self.batch_size[2]/len(intervals))
        leftovers = self.batch_size[2] - len(intervals)*n_selection_p_bin

        # select based on unlabeled U and y preds
        unlabeled_y_preds = pd.DataFrame(self._Y_pred[self.U_indices], columns=['yp'])

        i_dsa_queries = []
        if n_selection_p_bin > 0:
            for i, bin in enumerate(intervals):
                indices = list(unlabeled_y_preds[(unlabeled_y_preds['yp'] > bin.left) & (unlabeled_y_preds['yp'] <= bin.right)].index)
                if len(indices) > n_selection_p_bin:
                    select = list(np.random.choice(indices, n_selection_p_bin, replace=False))
                else:
                    select = indices
                    leftovers += n_selection_p_bin - len(indices)
                i_dsa_queries += select

        # leftovers
        n = 0
        while len(i_dsa_queries) < self.batch_size[2]:
            bin = intervals[n]
            indices = list(unlabeled_y_preds[(unlabeled_y_preds['yp'] > bin.left) & (unlabeled_y_preds['yp'] <= bin.right)].index)
            if len(indices)> n_selection_p_bin+int(leftovers/len(intervals)) :
                select = np.random.choice(indices, 1)[0]
                while select in i_dsa_queries:
                    select = np.random.choice(indices, 1)[0]
            else:
                if n == len(intervals) - 1:
                    break
                else:
                    n+=1
                    continue
            i_dsa_queries.append(select)
            if n == len(intervals)-1:
                n=0
            else:
                n+=1

        assert len(i_dsa_queries) <= self.batch_size[2]

        return i_dsa_queries

    def _dsa_test_y(self):
        """
        dsa approach based on the test data and predicted y values
        """
        self._history_update(self._Y_pred.reshape(-1,))
        uncertainty_change = self._uncertainty_tracker(self._history[:,-1], self.history-1)      #shape: (m,)
        uncertainty_change_test = uncertainty_change[self.test_indices]
        votes = pd.DataFrame(uncertainty_change_test, columns=['uc'])
        votes['ind'] = votes.index
        votes.sort_values('uc', ascending=False, inplace=True)
        unlabeled_y_preds = self._Y_pred[self.U_indices]
        i_dsa_queries = []
        n = 1
        while len(i_dsa_queries) < self.batch_size[2]:
            select = list(votes.head(n)['ind'])
            idx = self.find_nearest(unlabeled_y_preds, self._Y_pred[self.test_indices[select[-1]]])
            n+=1
            if idx not in i_dsa_queries:
                i_dsa_queries.append(idx)
        return i_dsa_queries

    def find_nearest(self, array, value):
        array = array.reshape(-1,)
        idx = (np.abs(array - value)).argmin()
        return idx

    def _dsa_unlabeled(self,deviations):
        """
        dsa approach based on unlabeled data
        """
        dev = deviations.mean(axis=1)
        self._history_update(dev)
        uncertainty_change = self._uncertainty_tracker(self._history[:,-1], self.history-1)      #shape: (m,)
        uncertainty_change = uncertainty_change[self.U_indices]
        uncertainty_change = pd.DataFrame(uncertainty_change, columns=['uc'])
        uncertainty_change['ind'] = uncertainty_change.index
        uncertainty_change.sort_values('uc', ascending=False, inplace=True)
        i_dsa_queries = list(uncertainty_change.head(self.batch_size[2])['ind'])
        return i_dsa_queries

    def _uncertainty_tracker(self, dev, i):
        """
        recurcsive deviation of the history
        """
        i -= 1
        if i>= 0:
            dev = dev - self._history[:,i]
            dev = self._uncertainty_tracker(dev, i)
        return dev

    def _history_update(self, dev):
        """
        update last column of the history with dev and shift all the previous columns to the left.
        """
        for i in range(self.history-1):
            self._history[:,i] = self._history[:,i+1]
        self._history[:,-1] = dev

    def _bemcm(self, deviations, lin_layer, alpha, normalize_internal, former_queries):
        """
        bemcm approach
        B_EMCM = EMCM - correlation_term
        EMCM = mean(deviations * lin_layer)
        """
        former_queries = np.unique(former_queries)
        bemcm_size = sum(self.batch_size) - len(former_queries)
        # shapes: m = number of samples, d = length of latent features
        i_queries = []              # the indices of queries based on the length of U_indices (not original U)
        correlation_term = {it: np.zeros(lin_layer.shape) for it in range(deviations.shape[1])}   # shape of each: (m,d)
        while len(i_queries) < bemcm_size:
            norms = pd.DataFrame()  # store the norms of n_ensemble models
            for it in range(deviations.shape[1]):
                # EMCM
                EMCM = deviations[:,it].reshape(-1,1) * lin_layer

                # correlation term acts after first selection
                if len(i_queries) > 0:
                    correlation_term[it] += self._correlation_term(i_queries[-1], deviations[:,it], lin_layer, normalize_internal)

                # calculate batch EMCM
                B_EMCM = EMCM - alpha * correlation_term[it]    # shape: (m,d)

                # find norm and
                norm = np.linalg.norm(B_EMCM, axis=1)
                norms[it] = norm.reshape(-1,)

            # average and argmax of norms
            assert norms.shape == (self.U_indices.shape[0], deviations.shape[1])
            norms['mean'] = norms.mean(axis=1)
            norms['ind'] = norms.index
            norms = norms.drop(i_queries, axis=0)   # remove rows of previously selected points
            norms.sort_values('mean', ascending=False, inplace=True)

            # memorize the initial ranking of the norms
            if len(i_queries) == 0:
                initial_ranking = list(norms.head(bemcm_size)['ind'])

            # select top candidate and update i_queries
            n = 1
            select = list(norms.head(n)['ind'])
            while select[0] in former_queries:
                n += 1
                select = [list(norms.head(n)['ind'])[-1]]

            i_queries += select

        # make sure correlation term is making any difference than simple sorting of the initial norms
        if set(initial_ranking) <= set(i_queries):
            msg = "It seems that the correlation term is not effective. The initial ranking of the candidates are same as the final results."
            warnings.warn(msg)

        return i_queries

    def _correlation_term(self, ind, dev, lin_layer, normalize_internal):
        """
        The internal function to find the correlation term
        """
        deviations_asterisk = dev[ind] * lin_layer[ind]    # shape: (d,)
        deviations_asterisk_transpose = deviations_asterisk.reshape(-1,1).T  # shape (1,d)
        lin_layer_transpose = lin_layer.T       # shape: (d,m)
        dot = np.dot(deviations_asterisk_transpose, lin_layer_transpose)  # shape: (1,m)
        correlation_term_of_ind = dot.T * lin_layer       # shape (m,d) same shape as lin_layer
        if normalize_internal:
            scaler = StandardScaler()
            correlation_term_of_ind = scaler.fit_transform(correlation_term_of_ind)
        return correlation_term_of_ind

    def _train_predict_evaluate(self, model=None, data_list=None, Y_scaler=None, metrics=None, **kwargs):
        """
        This internal function trains the model, predicts the test values, calculate metrics if requested.

        Returns
        -------
        class
            trained model

        ndarray
            predicted values

        float
            MAE

        float
            RMSE

        float
            R-squared
        """
        model.fit(data_list[0], data_list[1], **kwargs)
        preds = model.predict(data_list[2])
        if Y_scaler is not None:
            preds = Y_scaler.inverse_transform(preds)

        mae=None; rmse=None; r2=None
        if isinstance(metrics, np.ndarray):
            mae = mean_absolute_error(metrics, preds)
            rmse = np.sqrt(mean_squared_error(metrics, preds))
            r2 = r2_score(metrics, preds)

        return model, preds, mae, rmse, r2

    def random_search(self, Y, test_type='passive', scale=True, n_evaluation=10, random_state=90, **kwargs):
        """
        This function randomly select same number of data points as the active learning rounds and store the results.

        Parameters
        ----------
        Y: array-like
            The 2-dimensional label for all the candidates in the pool. Basically, you won't run this method unless you have the labels
            for all your samples. Otherwise, trust us and perform an active learning search.

        test_type: str, optional (default = 'passive')
            The parameter value must be either 'passive' or 'active'.
            If passive, the initial randomly selected test set in the initialize method will be used for evaluation.
            If active, the current test set of active learning approach will be used for evaluation. Thus, if the test_type in active
            learning method is 'passive', you should run active and random search back to back and then deposit the data.
            This way you make sure both active and random search are tested on the same test sets.

        scale: bool or list, optional (default = True)
            if True, sklearn.preprocessing.StandardScaler will be used to scale X and Y before training.
            You can also pass a list of two scaler instances that perform sklearn-style fit_transform and transform methods
            for the X and Y, respectively.

        n_evaluation: int, optional (default = 3)
            number of times to repeat training of the model and evaluation on test set.

        random_state: int or RandomState, optional (default = 90)
            The random state will be directly passed to the sklearn.model_selection methods.
            Please find additional info at: https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.KFold.html

        kwargs
            Any argument (except input data) that should be passed to the model's fit method.

        Attributes
        ----------
        random_results: pandas dataframe
            The results from the random sampling to provide the baseline for your active learning search.
            You need to

        Notes
        -----
            - This method replicate the active learning training size with random sampling approach. Thus, you can run
            this function only if the results is not empty, i.e., you have run the active learning search at least once.

        """

        if self.query_number==0:
            msg = "You must run active learning search first."
            print(msg)
            return False
        # elif len(self._random_results)==0:
        #     self._random_results.append(self._results[0])

        Y = np.array(Y)
        if Y.shape[0] != self.U_size:
            msg = "The length of the Y array must be equal to the number of candidates in the U."
            print(msg)
            raise ValueError(msg)

        # find all indices except test indices
        # we must consider the test type: active or passive
        if test_type == 'passive':
            test_indices = self.initial_test_indices
            except_test_inds = [i for i in range(self.U_size) if i not in self.initial_test_indices]
        elif test_type == 'active':
            test_indices = self.test_indices
            except_test_inds = [i for i in range(self.U_size) if i not in self.test_indices]
        else:
            msg = "The parameter 'test_type' must be either 'passive' or 'active'."
            raise ValueError(msg)

        # remaining training set size to run ML
        remaining_ind = [i for i in range(len(self._results)) if i not in range(len(self._random_results))]
        for ind in remaining_ind:
            # get training set size
            tr_size = self._results[ind][1] # the index 1 must be number of training data

            # training and evaluation
            it_results = {'mae': [], 'rmse': [], 'r2': []}

            # shuffle split, n_evaluation times
            ss = ShuffleSplit(n_splits=1, test_size=None, train_size=tr_size,
                              random_state=random_state)
            for train_indices, _ in ss.split(except_test_inds):
                # training indices based on the original U
                actual_tr_inds = np.array(except_test_inds)[train_indices]
                X_tr = self.U[actual_tr_inds]
                Y_tr = Y[actual_tr_inds]

                # test set based on the test type
                X_te = self.U[test_indices]
                Y_te = Y[test_indices]
                # scale
                X_scaler, Y_scaler = self._scaler(scale)
                if X_scaler is not None:
                    # scale X arrays
                    _ = X_scaler.fit_transform(self.U)
                    X_tr = X_scaler.transform(X_tr)
                    X_te = X_scaler.transform(X_te)
                    # scale Y
                    Y_tr = Y_scaler.fit_transform(Y_tr)

                for it in range(n_evaluation):
                    model = self.model_creator()
                    model, Y_te_pred, mae, rmse, r2 = self._train_predict_evaluate(model, [X_tr, Y_tr, X_te],
                                                                                   Y_scaler,
                                                                                   Y_te,
                                                                                   **kwargs)
                    # metrics
                    it_results['mae'].append(mae)
                    it_results['rmse'].append(rmse)
                    it_results['r2'].append(r2)

                    # delete from memory
                    del model
                del X_tr, X_te, Y_tr
            # store evaluation results
            results_temp = [self._results[ind][0], tr_size, len(self.test_indices)]
            for metric in ['mae', 'rmse', 'r2']:
                results_temp.append(np.mean(np.array(it_results[metric]), axis=0))
                results_temp.append(np.std(np.array(it_results[metric]), axis=0, ddof=0))
            self._random_results.append(results_temp)

        return True

    def visualize(self, Y=None):
        """
        This function plot distribution of labels and principal components of the features for the last round of the
        active learning search.
        Note that this function uses the prediction values in the attribute `Y_pred`. This attribute will be updated after
        each round of search. Thus, we recommend you run `visualize` right after each call of search method to get a trajectory
        of the active learning process.

        Parameters
        ----------
        Y: array-like, optional (default = None)
            The 2-dimensional label for all the candidates in the pool (in case you have them!!!).
            If you have all the labels, we will be able to produce additional cool visualizations.


        Returns
        -------
        list
            A list of matplotlib.figure.Figure or tuples. This object contains information about the plot

        """
        if self.query_number == 0 and len(self._queries) > 0:
            msg = "Data is not available for visualization yet. You must deposit data first."
            raise ValueError(msg)

        collect_plots = {}

        # feature transformation
        pca = PCA(n_components=2)
        u = pca.fit_transform(self.U)   # use this (original) transformed feature space for all the X data
        u_rem = u[self.U_indices]
        # test is fixed
        xte = u[self.test_indices]
        # all trainings at the current state
        xtr = u[self.train_indices]
        if self.query_number == 0 or (self.query_number == 1 and len(self._queries) > 0):
            xtr_last_batch = u[self.train_indices][:self.train_size]
            ytr_last_batch = self._Y_train[:self.train_size]
        else:
            xtr_last_batch = u[self.train_indices][-sum(self.batch_size):]
            ytr_last_batch = self._Y_train[-sum(self.batch_size):]

        # plot1 : x/pc distribution
        collect_plots["dist_pc"] = self._visualize_dist_pc(u,u_rem, xte, xtr, xtr_last_batch)

        # plot2 : y distribution
        collect_plots["dist_y"] = self._visualize_dist_y(ytr_last_batch, Y)


        # plot3 : results' learning curves
        if len(self._results)>0:
            collect_plots["learning_curve"] = self._visualize_learning_curve()

        return collect_plots

    def _visualize_dist_pc(self, u, u_rem, xte, xtr, xtr_last_batch):
        """plot #1 : The distribution of PC1 values
        """

        import matplotlib
        matplotlib.use('Agg')
        import seaborn as sns
        import matplotlib.pyplot as plt

        # full
        fig1 = plt.figure()
        ax = sns.distplot(u[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["denim blue"],
                          kde_kws={"shade": True, 'bw': 0.15}, label='Original U')  # label='Original U',
        sns.distplot(u_rem[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["tangerine"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax, label='Current U')  # label='Current U',
        sns.distplot(xtr[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["medium green"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax, label='Entire Train')  # label='Entire Train',
        sns.distplot(xte[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["pale red"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax, label='Test')  # label='Entire Test',
        if sum(self.batch_size) != 1:
            sns.distplot(xtr_last_batch[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["light purple"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax, label='Last Batch')  # label='Last Batch',
        # labels
        ax.set_ylabel(r'$\pi(PC1)$')
        ax.set_xlabel(r'$Features-PC1$')

        # font size
        for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                     ax.get_xticklabels() + ax.get_yticklabels()):
            item.set_fontsize(14)
        
        
        # all except last batch
        fig2 = plt.figure()
        ax2 = sns.distplot(u[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["denim blue"],
                          kde_kws={"shade": True, 'bw': 0.15}, label='Original U')  # label='Original U',
        sns.distplot(u_rem[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["tangerine"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax2, label='Current U')  # label='Current U',
        sns.distplot(xtr[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["medium green"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax2, label='Entire Train')  # label='Entire Train',
        sns.distplot(xte[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["pale red"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax2, label='Test')  # label='Entire Test',

        # labels
        ax2.set_ylabel(r'$\pi(PC1)$')
        ax2.set_xlabel(r'$Features-PC1$')

        # font size
        for item in ([ax2.title, ax2.xaxis.label, ax2.yaxis.label] +
                     ax2.get_xticklabels() + ax2.get_yticklabels()):
            item.set_fontsize(14)


        # all except Us
        fig3 = plt.figure()
        ax3 = sns.distplot(xtr[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["medium green"],
                     kde_kws={"shade": True, 'bw': 0.15}, label='Entire Train')  # label='Entire Train',
        sns.distplot(xte[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["pale red"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax3, label='Test')  # label='Entire Test',
        if sum(self.batch_size) != 1:
            sns.distplot(xtr_last_batch[:, 0].reshape(-1, ), hist=False, color=sns.xkcd_rgb["light purple"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax3, label='Last Batch')  # label='Last Batch',

        # labels
        ax3.set_ylabel(r'$\pi(PC1)$')
        ax3.set_xlabel(r'$Features-PC1$')

        # font size
        for item in ([ax3.title, ax3.xaxis.label, ax3.yaxis.label] +
                     ax3.get_xticklabels() + ax3.get_yticklabels()):
            item.set_fontsize(14)

        return (fig1, fig2, fig3)

    def _visualize_dist_y(self, ytr_last_batch, Y=None):
        """plot #2 : The distribution of y values
        """

        import matplotlib
        matplotlib.use('Agg')
        import seaborn as sns
        import matplotlib.pyplot as plt

        # full
        fig1 = plt.figure()
        ax = sns.distplot(self._Y_pred.reshape(-1, ), hist=False, color=sns.xkcd_rgb["tangerine"],
                           kde_kws={"shade": True, 'bw': 0.15}, label='Predicted U')  # label='Predicted U',
        if Y is not None:
            sns.distplot(Y.reshape(-1, ), hist=False, color=sns.xkcd_rgb["denim blue"],
                         kde_kws={"shade": True, 'bw': 0.15}, ax=ax, label='Labeled U')  # label='Labeled U',
        sns.distplot(self._Y_train.reshape(-1, ), hist=False, color=sns.xkcd_rgb["medium green"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax, label='Entire Train')  # label='Train',
        sns.distplot(self._Y_test.reshape(-1, ), hist=False, color=sns.xkcd_rgb["pale red"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax, label='Test')  # label='Test',
        if sum(self.batch_size) != 1:
            sns.distplot(ytr_last_batch.reshape(-1, ), hist=False, color=sns.xkcd_rgb["light purple"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax, label='Last Batch')  # label='Last Batch',
        # labels
        ax.set_ylabel(r'$\pi(Labels)$')
        ax.set_xlabel(r'$Labels$')

        # font size
        for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                     ax.get_xticklabels() + ax.get_yticklabels()):
            item.set_fontsize(14)

        # full except last batch
        fig2 = plt.figure()
        ax2 = sns.distplot(self._Y_pred.reshape(-1, ), hist=False, color=sns.xkcd_rgb["tangerine"],
                           kde_kws={"shade": True, 'bw': 0.15}, label='Predicted U')  # label='Predicted U',
        if Y is not None:
            sns.distplot(Y.reshape(-1, ), hist=False, color=sns.xkcd_rgb["denim blue"],
                         kde_kws={"shade": True, 'bw': 0.15}, ax=ax2, label='Labeled U')  # label='Labeled U',
        sns.distplot(self._Y_train.reshape(-1, ), hist=False, color=sns.xkcd_rgb["medium green"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax2, label='Entire Train')  # label='Train',
        sns.distplot(self._Y_test.reshape(-1, ), hist=False, color=sns.xkcd_rgb["pale red"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax2, label='Test')  # label='Test',

        # labels
        ax2.set_ylabel(r'$\pi(Labels)$')
        ax2.set_xlabel(r'$Labels$')

        # font size
        for item in ([ax2.title, ax2.xaxis.label, ax2.yaxis.label] +
                     ax2.get_xticklabels() + ax2.get_yticklabels()):
            item.set_fontsize(14)

        # full except Us
        fig3 = plt.figure()
        ax3 = sns.distplot(self._Y_train.reshape(-1, ), hist=False, color=sns.xkcd_rgb["medium green"],
                     kde_kws={"shade": True, 'bw': 0.15}, label='Entire Train')  # label='Train',
        sns.distplot(self._Y_test.reshape(-1, ), hist=False, color=sns.xkcd_rgb["pale red"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax3, label='Test')  # label='Test',
        if sum(self.batch_size) != 1:
            sns.distplot(ytr_last_batch.reshape(-1, ), hist=False, color=sns.xkcd_rgb["light purple"],
                     kde_kws={"shade": True, 'bw': 0.15}, ax=ax3, label='Last Batch')  # label='Last Batch',
        # labels
        ax3.set_ylabel(r'$\pi(Labels)$')
        ax3.set_xlabel(r'$Labels$')

        # font size
        for item in ([ax3.title, ax3.xaxis.label, ax3.yaxis.label] +
                     ax3.get_xticklabels() + ax3.get_yticklabels()):
            item.set_fontsize(14)

        return (fig1, fig2, fig3)

    def _visualize_learning_curve(self):
        """plot #3 : The learning curve of results
        """

        import matplotlib
        matplotlib.use('Agg')
        import seaborn as sns
        import matplotlib.pyplot as plt

        # data preparation
        algorithm = ['EMC'] * 3 * len(self._results)
        mae = list(self.results['mae'])
        mae += list(self.results['mae'] + self.results['mae_std'])
        mae += list(self.results['mae'] - self.results['mae_std'])
        size = list(self.results['num_training']) * 3
        if len(self._random_results) > 0 :
            pad_size = len(self._results) - len(self._random_results)
            algorithm += ['Random'] * 3 * len(self._results)
            mae += list(self.random_results['mae'])
            mae += list(self.random_results['mae'] + self.random_results['mae_std']) + [0] * pad_size
            mae += list(self.random_results['mae'] - self.random_results['mae_std']) + [0] * pad_size
            size += list(self.results['num_training']) * 3
        # dataframe
        dp = pd.DataFrame()
        dp['Algorithm'] = algorithm
        dp['Mean Absolute Error'] = mae
        dp['Training Size'] = size

        # figures
        sns.set_style('whitegrid')
        fig = plt.figure()
        ax = sns.lineplot(x='Training Size',
                          y='Mean Absolute Error',
                          style='Algorithm',
                          hue='Algorithm',
                          markers={'EMC': 'o', 'Random': 's'},
                          palette={'EMC': sns.xkcd_rgb["denim blue"],
                                   'Random': sns.xkcd_rgb["pale red"]},
                          data=dp,
                          )

        # font size
        for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                     ax.get_xticklabels() + ax.get_yticklabels()):
            item.set_fontsize(14)

        return fig


#py2 and py3
from six import iteritems

from itertools import combinations as comb
import numpy as np
from ....utility.EqualSumCombinations import EqualSumCombinations
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData

class PhaseDiagramCompositionEntryGenerator:
    """Class to generate composition entries at many points in many phase
    diagrams. Has two different ways of selecting compositions within phase
    diagrams.
    1. Even Spacing: Compositions are selected to be evenly-spaced throughout
    the phase diagram (e.g. A0.2B0.2C0.6, A0.4B0.2C0.4 etc.). This method is
    most appropriate for alloy systems.
    2. Simple Fractions: Compositions with the smallest denominator are
    selected (e.g. ABC, A2C, B2C, etc.). This method is most appropriate for
    phase diagrams that represent ordered crystalline compounds.

    Attributes
    ----------
    e_ids : array-like
        A list of int values denoting the list of elements to use.
    min_order : int
        Minimum number of constituents.
    max_order : int
        Maximum number of constituents.
    even_spacing : bool
        Whether to use even spacing or small integers.
    size : int
        Either number of stops in each direction or max denominator.

    """

    # List of elements to use (id is Z-1).
    e_ids = None

    # Minimum number of constituents.
    min_order = 1

    # Maximum number of constituents.
    max_order = 1

    # Whether to use even spacing or small integers.
    even_spacing = True

    # Either number of stops in each direction or max denominator.
    size = 3

    def set_elements_by_index(self, indices):
        """Function to define the list of elements to be included in the phase
        diagrams.

        Parameters
        ----------
        indices : array-like
            List of elements by index (Z-1).

        Raises
        ------
        ValueError
            If an element index is out of the allowed range.

        """
        self.e_ids = []
        for i in indices:
            if i < len(LookUpData.element_names):
                self.e_ids.append(i)
            else:
                raise ValueError("Index out of range: "+str(i))

    def set_elements_by_name(self, names):
        """Function to define the list of elements to be included in the phase
        diagrams.

        Parameters
        ----------
        names : array-like
            List of element names.

        Raises
        ------
        ValueError
            If an element name is not valid.

        """

        self.e_ids = []
        for name in names:
            if name in LookUpData.element_names:
                self.e_ids.append(LookUpData.element_names.index(name))
            else:
                raise ValueError("Element: "+name+" invalid!")

    def set_order(self, min_, max_):
        """Function to define the order of generated phase diagrams.

        Parameters
        ----------
        min_ : int
            Minimum number of constituents.
        max_ : int
            Maximum number of constituents.

        Raises
        ------
        ValueError
            If min_ or max_ is less than 1.

        """
        if min_ < 1 or max_ < 1:
            raise ValueError("Orders must be greater than 1.")
        self.min_order = min_
        self.max_order = max_

    def set_even_spacing(self, es):
        """Function to define whether to use evenly-spaced compositions (or
        low-denominator).

        Parameters
        ----------
        es : bool
            Boolean indicating the same.

        """
        self.even_spacing = es

    def set_size(self, size):
        """Function to define the number of points per binary diagram (in using
        even spacing) or the maximum denominator (for low-denominator).

        Parameters
        ----------
        size : int
            Desired size parameter.

        Raises
        ------
        ValueError
            If input is less than 2.

        """
        if size < 2:
            raise ValueError("Size must be greater than 1.")
        self.size = size

    def generate_alloy_compositions(self):
        """Function to generate evenly-spaced compositions. Generates
        compositions for all diagrams up to the user-specified Minimum order.
        For example: If the user wants ternary diagrams with a minimum
        spacing of 0.25 this code will generate the following map:
        1 -> ([1.0])
        2 -> ([0.25, 0.75], [0.5, 0.5], [0.75, 0.25])
        3 -> ([0.5, 0.25, 0.25], [0.25, 0.5, 0.25], [0.25, 0.25, 0.5])

        Returns
        -------
        output : dict
            A dictionary containing <order, possible compositions> as <key,
            value> pairs. Here, order is the number of elements and possible
            compositions is a list of numpy arrays containing the fractions
            of elements.

        """

        output = {}

        # Add in diagrams of greater order.
        for order in range(self.min_order, self.max_order+1):
            if order == 1:
                tmp_list = []
                tmp_list.append(np.array([1.0]))
                output[order] = tmp_list
                continue

            tmp_list = []
            es = EqualSumCombinations(self.size-1, order)
            for compI in es.get_combinations(self.size - 1, order):
                if 0 in compI:
                    # Don't add compositions from a lower-order diagram.
                    continue
                comp = np.zeros(order)
                for i in range(order):
                    comp[i] = compI[i]/ float(self.size - 1.0)
                tmp_list.append(comp)
            output[order] = tmp_list
        return output

    def generate_crystal_compositions(self):
        """Function to generate compositions with small denominators. Generates
        compositions for all diagrams up to the user-specified Minimum order.
        For example: If the user wants ternary diagrams with a maximum
        denominator of 3 this code will generate the following map:
        1 -> ([1])
        2 -> ([1/3, 2/3], [1/2, 1/2], [2/3, 1/3])
        3 -> ([1/3, 1/3, 1/3])

        Returns
        -------
        output : dict
            A dictionary containing <order, possible compositions> as <key,
            value> pairs. Here, order is the number of elements and possible
            compositions is a list of numpy arrays containing the fractions
            of elements.

        """

        output = {}

        # Add in diagrams of greater order.
        for order in range(self.min_order, self.max_order+1):
            if order == 1:
                tmp_list = []
                tmp_list.append(np.array([1.0]))
                output[order] = tmp_list
                continue

            tmp_list = []
            reduced_examples = []
            for d in range(order, self.size+1):
                es = EqualSumCombinations(d, order)
                for compI in es.get_combinations(d, order):
                    if 0 in compI:
                        # Don't add compositions from a lower-order diagram.
                        continue
                    comp = np.zeros(order)
                    red_comp = np.zeros(order)
                    for i in range(order):
                        comp[i] = float(compI[i])
                        red_comp[i] = comp[i] / d

                    # Check if this composition is already represented.
                    was_found = False
                    for ex in reduced_examples:
                        if np.array_equal(red_comp, ex):
                            was_found = True
                            break

                    if not was_found:
                        tmp_list.append(comp)
                        reduced_examples.append(red_comp)
            output[order] = tmp_list
        return output

    def generate_entries(self):
        """Function to generate the list of entries corresponding to the list of
        compositions, element names specified by the user and the mapping of
        number of elements to compositions.

        Returns
        -------
        entries: array-like
            A list of CompositionEntry's.

        """
        entries = []

        # Get the correct composition mapping.
        compositions = self.generate_alloy_compositions() if \
            self.even_spacing else self.generate_crystal_compositions()

        for (order,list_of_fractions) in iteritems(compositions):

            # Generate all possible combinations of elements of a given order.
            compounds = [list(i) for i in comb(self.e_ids, order)]
            # print order, len(compounds)*len(list_of_fractions)
            for frac in list_of_fractions:
                for compound in compounds:
                    entry = CompositionEntry(element_ids=compound,
                                             fractions=list(frac))
                    entries.append(entry)
        return entries



from __future__ import print_function
from builtins import range
import pandas as pd
import numpy as np
import scipy.sparse


from chemml.chem import Molecule

class RDKitFingerprint(object):
    """
    This is an interface to the available molecular fingerprints in the RDKit package.

    Parameters
    ----------
    fingerprint_type : str, optional (default='Morgan')
        The type of fingerprint. Available fingerprint types:
            - 'hashed_atom_pair' or 'hap'
            - 'MACCS' or 'maccs'
            - 'morgan'
            - 'hashed_topological_torsion' or 'htt'
            - 'topological_torsion' or 'tt'

    vector : str, optional (default = 'bit')
        Available options for vector:
            - 'int' : represent counts for each fragment instead of bits
                    It is not available for 'MACCS'.
            - 'bit' : only zeros and ones
                    It is not available for 'Topological_torsion'.

    n_bits : int, optional (default = 1024)
        It sets number of elements/bits in the 'bit' type of fingerprint vectors.
        Not availble for:
            - 'MACCS' - (MACCS keys have a fixed length of 167 bits)
            - 'Topological_torsion' - doesn't return a bit vector at all.

    radius : int, optional (default = 2)
        only applicable if calculating 'Morgan' fingerprint.

    kwargs :
        Any additional argument that should be passed to the rdkit fingerprint function.

    Attributes
    ----------
    n_molecules_ : int
        The number of molecules that are received.

    fps_ : list
        The list of rdkit fingerprint objects.

    """

    def __init__(self,
                 fingerprint_type='Morgan',
                 vector='bit',
                 n_bits=1024,
                 radius=2,
                 **kwargs):
        self.fingerprint_type = fingerprint_type
        self.n_bits = n_bits
        self.radius = radius
        self.kwargs = kwargs
        if not isinstance(vector, str) or vector.lower() not in ('bit', 'int'):
            msg = "The parameter vector must be either 'int' or 'bit'."
            raise ValueError(msg)
        else:
            self.vector = vector.lower()

    def represent(self, molecules):
        """
        The main function to provide fingerprint representation of input molecule(s).

        Parameters
        ----------
        molecules : chemml.chem.Molecule object or list
            It must be an instance of chemml.chem.Molecule object or a list of those objects, otherwise a ValueError will be raised.
            If smiles representation of the molecule (or rdkit molecule object) is not available, we convert the molecule to
            smiles automatically. However, the automatic conversion may ignore your manual settings, for example removed hydrogens,
            kekulized, or canonical smiles.

        Returns
        -------
        features : pandas.DataFrame
            A 2-dimensional pandas dataframe of fingerprint features with same number of rows as number of molecules.

        """
        if isinstance(molecules, list):
            molecules = np.array(molecules)
        elif isinstance(molecules, Molecule):
            molecules = np.array([molecules])
        else:
            msg = "The molecule must be a chemml.chem.Molecule object or a list of objets."
            raise ValueError(msg)

        if molecules.ndim >1:
            msg = "The molecule must be a chemml.chem.Molecule object or a list of objets."
            raise ValueError(msg)

        self.n_molecules_ = molecules.shape[0]

        if self.fingerprint_type.lower() == 'hashed_atom_pair' or self.fingerprint_type.lower() == 'hap':
            return self._hap(molecules)
        elif self.fingerprint_type == 'MACCS' or self.fingerprint_type.lower() == 'maccs':
            return self._maccs(molecules)
        elif self.fingerprint_type.lower() == 'morgan':
            return self._morgan(molecules)
        elif self.fingerprint_type.lower() == 'hashed_topological_torsion' or self.fingerprint_type.lower() == 'htt':
            return self._htt(molecules)
        elif self.fingerprint_type.lower() == 'topological_torsion' or self.fingerprint_type.lower() == 'tt':
            return self._tt(molecules)
        else:
            msg = "The parameter 'fingerprint_type' is not a valid fingerprint type: '%s'" % self.fingerprint_type
            raise ValueError(msg)

    def _hap(self, molecules):
        if self.vector == 'int':
            from rdkit.Chem.AtomPairs.Pairs import GetHashedAtomPairFingerprint
            self.fps_ = [
                GetHashedAtomPairFingerprint(self._sanitary(m), nBits=self.n_bits, **self.kwargs)
                for m in molecules
            ]
            # get nonzero elements as a dictionary for each molecule
            dict_nonzero = [fp.GetNonzeroElements() for fp in self.fps_]
            data = pd.DataFrame(dict_nonzero)
            data.fillna(0, inplace=True)
            return data
        elif self.vector == 'bit':
            from rdkit.Chem.rdMolDescriptors import GetHashedAtomPairFingerprintAsBitVect
            self.fps_ = [
                GetHashedAtomPairFingerprintAsBitVect(
                    self._sanitary(m), nBits=self.n_bits, **self.kwargs) for m in molecules
            ]
            data = np.array(self.fps_)
            data = pd.DataFrame(data)
            return data

    def _maccs(self, molecules):
        if self.vector == 'int':
            msg = "There is no RDKit function to encode integer vectors for MACCS keys"
            raise ValueError(msg)
        elif self.vector == 'bit':
            from rdkit.Chem.MACCSkeys import GenMACCSKeys
            self.fps_ = [GenMACCSKeys(self._sanitary(mol), **self.kwargs) for mol in molecules]
            data = np.array(self.fps_)
            data = pd.DataFrame(data)
            return data

    def _morgan(self, molecules):
        if self.vector == 'int':
            from rdkit.Chem.rdMolDescriptors import GetMorganFingerprint
            self.fps_ = [
                GetMorganFingerprint(self._sanitary(mol), self.radius, **self.kwargs)
                for mol in molecules
            ]
            # get nonzero elements as a dictionary for each molecule
            dict_nonzero = [fp.GetNonzeroElements() for fp in self.fps_]
            # pairScores = []
            # for fp in dict_nonzero:
            #     pairScores += list(fp)
            data = pd.DataFrame(dict_nonzero)#, columns=list(set(pairScores)))
            data.fillna(0, inplace=True)
            return data
        elif self.vector == 'bit':
            from rdkit.Chem.rdMolDescriptors import GetMorganFingerprintAsBitVect
            self.fps_ = [
                GetMorganFingerprintAsBitVect(
                    self._sanitary(mol), self.radius, nBits=self.n_bits, **self.kwargs)
                for mol in molecules
            ]
            data = np.array(self.fps_)
            data = pd.DataFrame(data)
            return data

    def _htt(self, molecules):
        if self.vector == 'int':
            from rdkit.Chem.rdMolDescriptors import GetHashedTopologicalTorsionFingerprint
            self.fps_ = [
                GetHashedTopologicalTorsionFingerprint(
                    self._sanitary(mol), nBits=self.n_bits, **self.kwargs) for mol in molecules
            ]
            # get nonzero elements as a dictionary for each molecule
            dict_nonzero = [fp.GetNonzeroElements() for fp in self.fps_]
            data = pd.DataFrame(dict_nonzero)
            data.fillna(0, inplace=True)
            return data
        elif self.vector == 'bit':
            from rdkit.Chem.rdMolDescriptors import GetHashedTopologicalTorsionFingerprintAsBitVect
            self.fps_ = [
                GetHashedTopologicalTorsionFingerprintAsBitVect(
                    self._sanitary(mol), nBits=self.n_bits, **self.kwargs) for mol in molecules
            ]
            data = np.array(self.fps_)
            data = pd.DataFrame(data)
            return data

    def _tt(self, molecules):
        if self.vector == 'int':
            from rdkit.Chem.AtomPairs.Torsions import GetTopologicalTorsionFingerprintAsIntVect
            self.fps_ = [
                GetTopologicalTorsionFingerprintAsIntVect(self._sanitary(mol), **self.kwargs)
                for mol in molecules
            ]
            dict_nonzero = [fp.GetNonzeroElements() for fp in self.fps_]
            data = pd.DataFrame(dict_nonzero)
            data.fillna(0, inplace=True)
            return data
        elif self.vector == 'bit':
            msg = "There is no RDKit function to encode bit vectors for Topological Torsion Fingerprints"
            raise ValueError(msg)

    def _sanitary(self, mol):
        if not isinstance(mol, Molecule):
            msg = "The molecule must be a chemml.chem.Molecule object or a list of objets."
            raise ValueError(msg)
        if mol.rdkit_molecule is None:
            mol.to_smiles()
        return mol.rdkit_molecule

    def store_sparse(self, file, features):
        """
        This function helps you to store higly sparse fingerprint feature sets using `.npz` format for memory efficiency and
        less store/load time.
        Another method of this class, `load_sparse`, enables you to load your `.npz` files and convert it back to pandas dataframe.

        Parameters
        ----------
        file : str
            Must be a path to the file with .npz format.

        features : pandas DataFrame
            Must be the pandas dataframe as you receive it from `represent` method.

        """
        if not isinstance(file, str):
            msg = "The parameter 'file' must be a path to the file with .npz format."
            raise ValueError(msg)

        if not isinstance(features, pd.DataFrame):
            msg = "The parameter 'features' must be a pandas dataframe."
            raise ValueError(msg)

        temp = scipy.sparse.csc_matrix(features.values)
        scipy.sparse.save_npz(file, temp)

    def load_sparse(self, file):
        """
        This function enables you to load sparse matrix with the `.npz` format and convert it to a pandas dataframe.

        Parameters
        ----------
        file : str
            Must be a path to the file with .npz format.

        Returns
        -------
        features : pandas.DataFrame
            The dense dataframe of the passed sparse file.

        """
        if not isinstance(file, str):
            msg = "The parameter 'file' must be a path to the file with .npz format."
            raise ValueError(msg)

        temp = scipy.sparse.load_npz(file)
        return pd.DataFrame(temp.todense())



"""
The 'chemml.models' module includes (please click on links adjacent to function names for more information):
    - OrganicLorentzLorenz: :func:`~chemml.models.keras.trained.OrganicLorentzLorenz`
    - MLP: :func:`~chemml.models.keras.mlp.MLP`
"""

from chemml.models.keras.mlp import MLP

from chemml.models.keras.graphconvlayers import NeuralGraphHidden, NeuralGraphOutput

from chemml.models.keras.transfer import TransferLearning


__all__ = [
    'MLP',
    'NeuralGraphHidden',
    'NeuralGraphOutput',
    'TransferLearning'
]



from __future__ import print_function
import os
from tensorflow import keras # required to go around the protobuf error after importing pybel prior to tensorflow
from rdkit import Chem
from openbabel import pybel
from rdkit.Chem import AllChem
import warnings
import numpy as np

from ..utils import update_default_kwargs


class XYZ(object):
    """
    This class stores the information that is typically carried by standard XYZ files.

    Parameters
    ----------
    geometry : ndarray
        The numpy array of shape (number_of_atoms, 3).
        It stores the xyz coordinates for each atom of the molecule.

    atomic_numbers : ndarray
        The numpy array of shape (number_of_atoms, 1).
        It stores the atomic numbers of each atom in the molecule (in the same order as geometry).

    atomic_symbols : ndarray
        The numpy array of shape (number_of_atoms, 1).
        It stores the atomic symbols of each atom in the molecule (in the same order as geometry).

    """
    def __init__(self,geometry, atomic_numbers, atomic_symbols):
        # check data type
        if not (isinstance(geometry, np.ndarray) and isinstance(atomic_numbers,np.ndarray)
                and isinstance(atomic_symbols, np.ndarray)):
            msg = "The parameters' value must be numpy array."
            raise ValueError(msg)
        else:
            # check dimensions
            if geometry.ndim != 2 or atomic_symbols.ndim != 2 or atomic_numbers.ndim != 2:
                msg = "The parameters' arrays must be 2 dimensional."
                raise ValueError(msg)
        # cast data format
        try:
            geometry.astype(float)
            atomic_numbers.astype(int)
            atomic_symbols.astype(str)
        except ValueError:
            msg = "The input data must be able to be casted to the following specified types:\n" \
                  "    - geoometry      >> float\n" \
                  "    - atomic_numbers >> integer\n" \
                  "    - atomic_symbols >> string"
            raise ValueError(msg)
        # check shapes
        if geometry.shape[1] == 3:
            if geometry.shape[0] == atomic_symbols.shape[0] and \
               geometry.shape[0] == atomic_numbers.shape[0] :
                self.geometry = geometry
                self.atomic_numbers = atomic_numbers
                self.atomic_symbols = atomic_symbols
            else:
                msg = "The number of atoms in the molecule is different between entries."
                raise ValueError(msg)
        else:
            msg = "All three coordinate positions are required in the geometries."
            raise ValueError(msg)

    def __repr__(self):
        return '<XYZ(geometry: {self.geometry.shape!r},' \
               ' atomic_numbers: {self.atomic_numbers.shape!r},' \
               ' atomic_symbols: {self.atomic_symbols.shape!r})>'.format(self=self)


class Molecule(object):
    """
    The central class to construct a molecule from different chemical input formats.
    This module is built on top of RDKit and OpenBabel python API.
    We join the forces and strength of these two cheminformatic libraris for a consistent user experience.

    Almost all the molecular descriptors and molecule-based ML models require the chemical informatin as a Molecule object.
    Several methods are available in this module to facilitate the manipulation of chemical data.

    Parameters
    ----------
    input : str
        The representation string or path to a file.

    input_type : str
        The input type.
        The available types are enlisted here:
            - smiles: The input must be SMILES representation of a molecule.
            - smarts: The input must be SMARTS representation of a molecule.
            - inchi: The input must be InChi representation of a molecule.
            - xyz:  The input must be the path to an xyz file.

    kwargs :
        The corresponding RDKit arguments for each of the input types:
            - smiles: http://rdkit.org/docs/source/rdkit.Chem.rdmolfiles.html#rdkit.Chem.rdmolfiles.MolFromSmiles
            - smarts: http://rdkit.org/docs/source/rdkit.Chem.rdmolfiles.html#rdkit.Chem.rdmolfiles.MolFromSmarts
            - inchi: http://rdkit.org/docs/source/rdkit.Chem.inchi.html?highlight=inchi#rdkit.Chem.inchi.MolFromInchi

    Notes
    -----
        - The molecule will be creatd as an RDKit molecule object.
        - The molecule object will be stored and available as `rdkit_molecule` attribute.
        - If you load a molecule from its SMARTS string, there is high probability that you can't convert it to other
            types due to the abstract description of the molecules by SMARTS representation.

    Attributes
    ----------
    rdkit_molecule : object
        The `rdkit.Chem.rdchem.Mol` object

    smiles : str
        The SMILES string that you get by running the `to_smiles` method.

    smarts : str
        The SMARTS string that you get by running the `to_smarts` method.

    inchi : str
        The InChi string that you get by running the `to_inchi` method.

    xyz : instance of <class 'chemml.chem.molecule.XYZ'>
        The class object that stores the 3D info. The available attributes in the class are 'geometry',
        'atomic_numbers', and 'atomic_symbols'.

    Examples
    --------
    >>> from chemml.chem import Molecule
    >>> caffeine_smiles = 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'
    >>> caffeine_smarts = '[#6]-[#7]1:[#6]:[#7]:[#6]2:[#6]:1:[#6](=[#8]):[#7](:[#6](=[#8]):[#7]:2-[#6])-[#6]'
    >>> caffeine_inchi = 'InChI=1S/C8H10N4O2/c1-10-4-9-6-5(10)7(13)12(3)8(14)11(6)2/h4H,1-3H3'
    >>> mol = Molecule(caffeine_smiles, 'smiles')
    >>> mol
    <Molecule(rdkit_molecule : <rdkit.Chem.rdchem.Mol object at 0x11060a8a0>,
          creator        : ('SMILES', 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'),
          smiles         : 'Cn1c(=O)c2c(ncn2C)n(C)c1=O',
          smarts         : None,
          inchi          : None,
          xyz            : None)>
    >>> mol.smiles # this is the canonical SMILES by default
    'Cn1c(=O)c2c(ncn2C)n(C)c1=O'
    >>> mol.creator
    ('SMILES', 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C')
    >>> mol.to_smiles(kekuleSmiles=True)
    >>> mol
    <Molecule(rdkit_molecule : <rdkit.Chem.rdchem.Mol object at 0x11060a8a0>,
          creator        : ('SMILES', 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'),
          smiles         : 'CN1C(=O)C2=C(N=CN2C)N(C)C1=O',
          smarts         : None,
          inchi          : None,
          xyz            : None)>
    >>> mol.smiles # the kukule SMILES is not canonical
    'CN1C(=O)C2=C(N=CN2C)N(C)C1=O'
    >>> mol.inchi is None
    True
    >>> mol.to_inchi()
    >>> mol
    <Molecule(rdkit_molecule : <rdkit.Chem.rdchem.Mol object at 0x11060a8a0>,
          creator        : ('SMILES', 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'),
          smiles         : 'CN1C(=O)C2=C(N=CN2C)N(C)C1=O',
          smarts         : None,
          inchi          : 'InChI=1S/C8H10N4O2/c1-10-4-9-6-5(10)7(13)12(3)8(14)11(6)2/h4H,1-3H3',
          xyz            : None)>
    >>> mol.inchi
    'InChI=1S/C8H10N4O2/c1-10-4-9-6-5(10)7(13)12(3)8(14)11(6)2/h4H,1-3H3'
    >>> mol.to_smarts(isomericSmiles=False)
    >>> mol.smarts
    '[#6]-[#7]1:[#6]:[#7]:[#6]2:[#6]:1:[#6](=[#8]):[#7](:[#6](=[#8]):[#7]:2-[#6])-[#6]'
    >>>
    >>> # add hydrogens and recreate smiles and inchi
    >>> mol.hydrogens('add')
    >>> mol.to_smiles()
    >>> mol.to_inchi()
    <Molecule(rdkit_molecule : <rdkit.Chem.rdchem.Mol object at 0x11060ab70>,
          creator        : ('SMILES', 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'),
          smiles         : '[H]c1nc2c(c(=O)n(C([H])([H])[H])c(=O)n2C([H])([H])[H])n1C([H])([H])[H]',
          smarts         : '[#6]-[#7]1:[#6]:[#7]:[#6]2:[#6]:1:[#6](=[#8]):[#7](:[#6](=[#8]):[#7]:2-[#6])-[#6]',
          inchi          : 'InChI=1S/C8H10N4O2/c1-10-4-9-6-5(10)7(13)12(3)8(14)11(6)2/h4H,1-3H3',
          xyz            : None)>
    >>> # note that by addition of hydrogens, the smiles string changed, but not inchi
    >>> mol.to_xyz()
    ValueError: The conformation has not been built yet. Maybe due to the 2D representation of the creator.
    You should set the optimizer value if you wish to embed and optimize the 3D geometries.
    >>> mol.to_xyz('MMFF', maxIters=300, mmffVariant='MMFF94s')
    >>> mol
    <Molecule(rdkit_molecule : <rdkit.Chem.rdchem.Mol object at 0x11060ab70>,
          creator        : ('SMILES', 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'),
          smiles         : '[H]c1nc2c(c(=O)n(C([H])([H])[H])c(=O)n2C([H])([H])[H])n1C([H])([H])[H]',
          smarts         : '[#6]-[#7]1:[#6]:[#7]:[#6]2:[#6]:1:[#6](=[#8]):[#7](:[#6](=[#8]):[#7]:2-[#6])-[#6]',
          inchi          : 'InChI=1S/C8H10N4O2/c1-10-4-9-6-5(10)7(13)12(3)8(14)11(6)2/h4H,1-3H3',
          xyz            : <chemml.chem.molecule.XYZ object at 0x1105f34e0>)>
    >>> mol.xyz
    <chemml.chem.molecule.XYZ object at 0x1105f34e0>
    >>> mol.xyz.geometry
    array([[-3.13498321,  1.08078307,  0.33372515],
       [-2.15703638,  0.0494926 ,  0.0969075 ],
       [-2.41850776, -1.27323453, -0.14538583],
       [-1.30933543, -1.96359887, -0.32234393],
       [-0.31298208, -1.04463051, -0.1870366 ],
       [-0.80033326,  0.20055608,  0.07079013],
       [ 0.02979071,  1.33923464,  0.25484381],
       [-0.41969338,  2.45755985,  0.48688003],
       [ 1.39083332,  1.04039479,  0.14253256],
       [ 1.93405927, -0.23430839, -0.12320318],
       [ 3.15509616, -0.39892779, -0.20417755],
       [ 1.03516373, -1.28860035, -0.28833489],
       [ 1.51247526, -2.63123373, -0.56438513],
       [ 2.34337825,  2.12037198,  0.31039958],
       [-3.03038469,  1.84426033, -0.44113804],
       [-2.95880047,  1.50667225,  1.32459816],
       [-4.13807215,  0.64857048,  0.29175711],
       [-3.4224011 , -1.6776074 , -0.18199971],
       [ 2.60349515, -2.67375289, -0.61674561],
       [ 1.10339164, -2.96582894, -1.52299068],
       [ 1.17455601, -3.30144289,  0.23239402],
       [ 2.94406381,  2.20916763, -0.60086251],
       [ 1.86156872,  3.07985237,  0.51337935],
       [ 3.01465788,  1.87625024,  1.14039627]])
    >>> mol.xyz.atomic_numbers
    array([[6],
       [7],
       [6],
       [7],
       [6],
       [6],
       [6],
       [8],
       [7],
       [6],
       [8],
       [7],
       [6],
       [6],
       [1],
       [1],
       [1],
       [1],
       [1],
       [1],
       [1],
       [1],
       [1],
       [1]])
    >>> mol.xyz.atomic_symbols
    array([['C'],
           ['N'],
           ['C'],
           ['N'],
           ['C'],
           ['C'],
           ['C'],
           ['O'],
           ['N'],
           ['C'],
           ['O'],
           ['N'],
           ['C'],
           ['C'],
           ['H'],
           ['H'],
           ['H'],
           ['H'],
           ['H'],
           ['H'],
           ['H'],
           ['H'],
           ['H'],
           ['H']], dtype='<U1')
    """
    def __init__(self, input, input_type, **kwargs):
        self.rdkit_molecule = None
        self.pybel_molecule = None
        self.creator = None
        self._init_attributes()
        self._extra_docs()
        self._load(input, input_type, **kwargs)

    def __repr__(self):
        return '<chemml.chem.Molecule(\n' \
               '        rdkit_molecule : {self.rdkit_molecule!r},\n' \
               '        pybel_molecule : {self.pybel_molecule!r},\n' \
               '        creator        : {self.creator!r},\n' \
               '        smiles         : {self.smiles!r},\n' \
               '        smarts         : {self.smarts!r},\n' \
               '        inchi          : {self.inchi!r},\n' \
               '        xyz            : {self.xyz!r})>'.format(self=self)

    def _init_attributes(self):
        # SMILES
        self._smiles = None
        self._smiles_args = None
        # SMARTS
        self._smarts = None
        self._smarts_args = None
        # Inchi
        self._inchi = None
        self._inchi_args = None
        # xyz
        self._xyz = None
        self._UFF_args = None
        self._MMFF_args = None
        # default arguments
        self._default_rdkit_smiles_args = {"isomericSmiles":True, "kekuleSmiles":False, "rootedAtAtom":-1, "canonical":True, "allBondsExplicit":False,
                "allHsExplicit":False} #"doRandom":False
        self._default_rdkit_smarts_args = {"isomericSmiles":True}
        self._default_rdkit_inchi_args = {"options":'', 'logLevel':None, 'treatWarningAsError':False}
        self._default_UFF_args = {'maxIters':200, 'vdwThresh':10.0, 'confId':-1, 'ignoreInterfragInteractions':True}
        self._default_MMFF_args = {'mmffVariant':'MMFF94', 'maxIters':200, 'nonBondedThresh':100.0, 'confId':-1, 'ignoreInterfragInteractions':True}

    def _extra_docs(self):
        # method: to_smiles
        self._to_smiles_core_names = ("rdkit.Chem.MolToSmarts",)
        self._to_smiles_core_docs = ("http://rdkit.org/docs/source/rdkit.Chem.inchi.html?highlight=inchi#rdkit.Chem.inchi.MolToInchi",)
        # method: to_smarts
        self._to_smarts_core_names = ("rdkit.Chem.MolToSmarts",)
        self._to_smarts_core_docs = ("http://rdkit.org/docs/source/rdkit.Chem.inchi.html?highlight=inchi#rdkit.Chem.inchi.MolToInchi",)
        # method: to_inchi
        self._to_inchi_core_names = ("rdkit.Chem.MolToInchi",)
        self._to_inchi_core_docs = ("http://rdkit.org/docs/source/rdkit.Chem.inchi.html?highlight=inchi#rdkit.Chem.inchi.MolToInchi",)
        # method: hydrogens
        self._hydrogens_core_names = ("rdkit.Chem.AddHs","rdkit.Chem.RemoveHs")
        self._hydrogens_core_docs = ("http://rdkit.org/docs/source/rdkit.Chem.rdmolops.html?highlight=addhs#rdkit.Chem.rdmolops.AddHs",
                                    "http://rdkit.org/docs/source/rdkit.Chem.rdmolops.html?highlight=addhs#rdkit.Chem.rdmolops.RemoveHs")
        #
        self._to_xyz_core_names = ("rdkit.Chem.AllChem.MMFFOptimizeMolecule","rdkit.Chem.AllChem.UFFOptimizeMolecule")
        self._to_xyz_core_docs =(
            "http://rdkit.org/docs/source/rdkit.Chem.rdForceFieldHelpers.html?highlight=mmff#rdkit.Chem.rdForceFieldHelpers.MMFFOptimizeMolecule",
            "http://rdkit.org/docs/source/rdkit.Chem.rdForceFieldHelpers.html?highlight=mmff#rdkit.Chem.rdForceFieldHelpers.UFFOptimizeMolecule"
        )

    @property
    def smiles(self):
        return self._smiles

    @property
    def smiles_args(self):
        return self._smiles_args

    @property
    def smarts(self):
        return self._smarts

    @property
    def smarts_args(self):
        return self._smarts_args

    @property
    def inchi(self):
        return self._inchi

    @property
    def inchi_args(self):
        return self._inchi_args

    @property
    def xyz(self):
        return self._xyz

    @property
    def UFF_args(self):
        return self._UFF_args

    @property
    def MMFF_args(self):
        return self._MMFF_args

    def _check_original_molecule(self):
        if self.rdkit_molecule is None and self.pybel_molecule is None:
            msg = "Neither rdkit nor pybel molecule object has been created yet."
            raise ValueError(msg)
        elif self.rdkit_molecule is None:
            return 'pybel'
        elif self.pybel_molecule is None:
            return 'rdkit'
        # in case both of them are available go with rdkit
        else:
            return 'rdkit'

    def hydrogens(self, action='add', **kwargs):
        """
        This function adds/removes hydrogens to/from a prebuilt molecule object.

        Parameters
        ----------
        action : str
            Either 'add' or 'remove', to add hydrogns or remove them from the rdkit molecule.

        kwargs :
            The arguments that can be passed to the rdkit functions:
            - `Chem.AddHs`: documentation at http://rdkit.org/docs/source/rdkit.Chem.rdmolops.html?highlight=addhs#rdkit.Chem.rdmolops.AddHs
            - `Chem.RemoveHs`: documentation at http://rdkit.org/docs/source/rdkit.Chem.rdmolops.html?highlight=addhs#rdkit.Chem.rdmolops.RemoveHs

        Notes
        -----
            - The rdkit or pybel molecule object must be created in advance.
            - Only rdkit or pybel molecule object will be modified in place.
            - If you remove hydrogens from molecules, the atomic 3D coordinates might not be accurate for the conversion to xyz representation.

        """

        # molecule must exist
        _ = self._check_original_molecule()

        if action == 'add':
            if self.pybel_molecule:
                self.pybel_molecule.addh()
            # Note: just if not elif
            if self.rdkit_molecule:
                self.rdkit_molecule = Chem.AddHs(self.rdkit_molecule, **kwargs)
        elif action == 'remove':
            if self.pybel_molecule:
                self.pybel_molecule.removeh()
            if self.rdkit_molecule:
                self.rdkit_molecule = Chem.RemoveHs(self.rdkit_molecule, **kwargs)
        else:
            raise ValueError("The parameter 'action' must be either of 'add' or 'remove'.")

    def _load(self, input, input_type, **kwargs):
        """
        The main internal function to load a molecule.
        """
        if input_type == 'xyz':
            self._load_pybel(input, input_type, **kwargs)
        elif input_type in ['smiles', 'smarts', 'inchi']:
            # print(input)
            self._load_rdkit(input, input_type, **kwargs)
        else:
            msg = "The input type '%s' is not supported." %input_type
            raise ValueError(msg)

    def _load_rdkit(self, input, input_type, from_load=True, **kwargs):
        """
        The internal function to load a molecule using rdkit engine.
        """

        if input_type == 'smiles':
            creator = ('SMILES', input)
            rdkit_mol = Chem.MolFromSmiles(input, **kwargs)
        elif input_type == 'smarts':
            creator = ('SMARTS', input)
            rdkit_mol = Chem.MolFromSmarts(input, **kwargs)
        elif input_type == 'inchi':
            creator = ('InChi', input)
            rdkit_mol = Chem.MolFromInchi(input, **kwargs)
        else:
            msg = "The input type '%s' is not available from 'rdkit' engine." % (input_type)
            raise ValueError(msg)

        # check for None values. Some error messages that may have already raised:
        # [16:59:24] SMILES Parse Error: syntax error for input: 'NotSMILES'
        # [12:18:01] Explicit valence for atom # 1 O greater than permitted
        # [12:20:41] Can't kekulize mol
        if rdkit_mol is None:
            msg = 'The input is not a legit %s string. Refere to the error message that was already displayed by RDKit library!'%creator[0]
            raise ValueError(msg)

        # if the molecule is already being created warn the user
        if self.rdkit_molecule and self.creator:
            msg = "The molecule was already built using %s, is now overwritten by %s" % (
            str(self.creator), str(creator))
            warnings.warn(msg)

        self.rdkit_molecule = rdkit_mol

        if from_load:
            self.creator = creator
            self._init_attributes()

        if self.creator[0]=='SMILES':
            self.to_smiles()
        elif self.creator[0]=='SMARTS':
            self.to_smarts()
        elif self.creator[0]=='InChi':
            self.to_inchi()

    def _load_pybel(self, input, input_type, **kwargs):
        """
        The internal function to load a molecule using rdkit engine.

        """
        # available variables
        pybel_mol = None
        creator = None

        if input_type == 'xyz':
            if os.path.isfile(input):
                creator = ('XYZ', input)
                gen = pybel.readfile("xyz", input)
                mols = list(gen)
                if len(mols) == 1:
                    pybel_mol = mols[0]
                else:
                    warnings.warn('More than one Molecule object is created and is returned as a generator.')
                    return self._multiple_molecules(mols, creator)
            else:
                msg = "The input '%s' is not a valid XYZ input file."%input
                raise ValueError(msg)

        if pybel_mol is None:
            msg = 'The input is not a legit %s string. Refere to the error message that was already displayed by Pybel library!' % \
              creator[0]
            raise ValueError(msg)

        # if the molecule is already being created warn the user
        if self.pybel_molecule and self.creator:
            msg = "The molecule was already built using %s, is now overwritten by %s" % (
            str(self.creator), str(creator))
            warnings.warn(msg)

        self.creator = creator
        self.pybel_molecule = pybel_mol

        self._init_attributes()

        if self.creator[0]=='XYZ':
            self.to_xyz()

    def _multiple_molecules(self, mols, creator):
        for mol in mols:
            m = Molecule()
            m.pybel_molecule = mol
            m.creator = creator
            yield m

    def to_smiles(self, **kwargs):
        """
        This function creates and stores the SMILES string for a pre-built molecule.

        Parameters
        ----------
        kwargs :
            The arguments for the rdkit.Chem.MolToSmiles function.
            The documentation is available at: http://rdkit.org/docs/source/rdkit.Chem.rdmolfiles.html#rdkit.Chem.rdmolfiles.MolToSmiles

        Notes
        -----
            - The rdkit or pybel molecule object must be created in advance.
            - If only pybel molecule is available, we create an rdkit molecule using its SMILES representation, and then recreate the SMILES string using rdkit arguments.
            - The molecule will be modified in place.
            - For rdkit molecule the SMILES string is canocical by default, unless when one requests kekuleSmiles.

        """
        # molecule must exist
        engine = self._check_original_molecule()
        if engine == 'pybel':
            smiles = self.pybel_molecule.write('smi').strip().split('\t')[0]
            self._load_rdkit(smiles, 'smiles', from_load=False)
            self.to_smiles()
        else:
            self._to_smiles_rdkit(**kwargs)

    def _to_smiles_rdkit(self, **kwargs):
        """
        This internal function creates and stores the SMILES string for rdkit molecule.
        """
        # kekulize flag
        if 'kekuleSmiles' in kwargs and kwargs['kekuleSmiles']:
            Chem.Kekulize(self.rdkit_molecule)

        # store arguments for future reference
        self._smiles = Chem.MolToSmiles(self.rdkit_molecule, **kwargs)

        # arguments
        self._smiles_args = update_default_kwargs(self._default_rdkit_smiles_args, kwargs,
                                                 self._to_smiles_core_names[0], self._to_smiles_core_docs[0])

    def to_smarts(self, **kwargs):
        """
        This function creates and stores the SMARTS string for a pre-built molecule.

        Parameters
        ----------
        kwargs :
            All the arguments that can be passed to the rdkit.Chem.MolToSmarts function.
            The documentation is available at: http://rdkit.org/docs/source/rdkit.Chem.rdmolfiles.html#rdkit.Chem.rdmolfiles.MolToSmarts

        Notes
        -----
            - The rdkit or pybel molecule object must be created in advance.
            - If only pybel molecule is available, we create an rdkit molecule using its SMILES representation, and then create the SMARTS string using rdkit arguments.
            - The molecule will be modified in place.

        """
        # molecule must exist
        engine = self._check_original_molecule()

        if engine == 'rdkit':
            self._to_smarts_rdkit(**kwargs)
        else:
            smiles = self.pybel_molecule.write('smi').strip().split('\t')[0]
            self._load_rdkit(smiles, 'smiles', from_load=False)
            self._to_smarts_rdkit(**kwargs)

    def _to_smarts_rdkit(self, **kwargs):
        """
        The internal
        """
        # store arguments for future reference
        self._smarts = Chem.MolToSmarts(self.rdkit_molecule, **kwargs)

        # arguments
        self._smarts_args = update_default_kwargs(self._default_rdkit_smarts_args, kwargs,
                                                 self._to_smarts_core_names[0], self._to_smarts_core_docs[0])

    def to_inchi(self, **kwargs):
        """
        This function creates and stores the InChi string for a pre-built molecule.

        Parameters
        ----------
        kwargs :
            The arguments that can be passed to the rdkit.Chem.MolToInchi function (will be used only if rdkit molecule is available).
            The documentation is available at: http://rdkit.org/docs/source/rdkit.Chem.inchi.html?highlight=inchi#rdkit.Chem.inchi.MolToInchi

        Notes
        -----
            - The rdkit or pybel molecule object must be created in advance.
            - The molecule will be modified in place.

        """
        # molecule must exist
        engine = self._check_original_molecule()
        if engine == 'pybel':
            self._to_inchi_pybel()
        else:
            self._to_inchi_rdkit(**kwargs)

    def _to_inchi_rdkit(self, **kwargs):
        """
        This internal function creates and stores the InChi string for rdkit molecule.
        """
        # store arguments for future reference
        self._inchi = Chem.MolToInchi(self.rdkit_molecule, **kwargs)

        # arguments
        self._inchi_args = update_default_kwargs(self._default_rdkit_inchi_args, kwargs,
                                                 self._to_inchi_core_names[0], self._to_inchi_core_docs[0])

    def _to_inchi_pybel(self):
        """
        This internal function creates and stores the InChi string for pybel molecule.
        """
        self._inchi = self.pybel_molecule.write('inchi').strip()

    def to_xyz(self, optimizer=None, **kwargs):
        """
        This function creates and stores the xyz coordinates for a pre-built molecule object.

        Parameters
        ----------
        optimizer : None or str, optional (default: None)
            If None, the geometries will be extracted from the available source of 3D structure (if any).
            Otherwise, any of the 'UFF' or 'MMFF' force fileds should be passed to embed and optimize geometries using 'rdkit.Chem.AllChem.UFFOptimizeMolecule' or
            'rdkit.Chem.AllChem.MMFFOptimizeMolecule' methods, respectively.

        kwargs :
            The arguments that can be passed to the corresponding forcefileds.
            The documentation is available at:
                - UFFOptimizeMolecule: http://rdkit.org/docs/source/rdkit.Chem.rdForceFieldHelpers.html?highlight=mmff#rdkit.Chem.rdForceFieldHelpers.UFFOptimizeMolecule
                - MMFFOptimizeMolecule: http://rdkit.org/docs/source/rdkit.Chem.rdForceFieldHelpers.html?highlight=mmff#rdkit.Chem.rdForceFieldHelpers.MMFFOptimizeMolecule

        Notes
        -----
            - The geometry will be stored in the `xyz` attribute.
            - The molecule object must be created in advance.
            - The hydrogens won't be added to the molecule automatically. You should add it manually using `hydrogens` method.
            - If the molecule object has been built using 2D representations (e.g., SMILES or InChi), the conformer
            doesn't exist and you nedd to set the optimizer parameter to any of the force fields.
            - If the 3D info exist but you still need to run optimization, the 3D structure will be embedded from scratch (i.e., the current atom coordinates will be removed.)


        """
        # rdkit molecule must exist
        engine = self._check_original_molecule()
        # any futher process depends on the state of the optimizer and available 3D info
        if self.xyz and optimizer is None:
            return True     # the required info is available
        elif self.pybel_molecule:
            if optimizer is None :
                self._to_xyz_pybel()    # pybel molecule always contains the 3D info
            else:
                # build the rdkit molecule from 2D info
                smiles = self.pybel_molecule.write('smi').strip().split('\t')[0]
                self._load_rdkit(smiles, 'smiles', from_load=False)
                self.creator = ('SMILES', smiles)
                self.hydrogens('add')
                self._to_xyz_rdkit(optimizer, **kwargs)
        elif engine == 'rdkit':
            self._to_xyz_rdkit(optimizer, **kwargs)

    def _to_xyz_rdkit(self, optimizer, **kwargs):
        """
        The internal function creates and stores the xyz coordinates for a pre-built molecule object.
        """
        # add hydrogens >> commented out and left for the users to take care of it using hydrogens method.
        # self.hydrogens('add')

        # embeding and optimization
        if optimizer:
            AllChem.EmbedMolecule(self.rdkit_molecule)
            if optimizer ==  'MMFF':
                if AllChem.MMFFHasAllMoleculeParams(self.rdkit_molecule):
                    AllChem.MMFFOptimizeMolecule(self.rdkit_molecule, **kwargs)
                else:
                    msg = "The MMFF parameters are not available for all of the molecule's atoms."
                    raise ValueError(msg)
            elif optimizer == 'UFF':
                if AllChem.UFFHasAllMoleculeParams(self.rdkit_molecule):
                    AllChem.UFFOptimizeMolecule(self.rdkit_molecule, **kwargs)
                else:
                    msg = "The UFF parameters are not available for all of the molecule's atoms."
                    raise ValueError(msg)
            else:
                msg = "The '%s' is not a legit value for the optimizer parameter."%str(optimizer)
                raise ValueError(msg)

        # get conformer
        try:
            conf = self.rdkit_molecule.GetConformer()
        except ValueError:
            msg = "The conformation has not been built yet (maybe due to the 2D representation of the creator).\n" \
                  "You should set the optimizer value if you wish to embed and optimize the 3D geometry."
            raise ValueError(msg)
        geometry = conf.GetPositions()
        atoms_list = self.rdkit_molecule.GetAtoms()
        atomic_nums = np.array([i.GetAtomicNum() for i in atoms_list])
        atomic_symbols = np.array([i.GetSymbol() for i in atoms_list])
        self._xyz = XYZ(geometry, atomic_nums.reshape(-1,1), atomic_symbols.reshape(-1,1))

        if optimizer=='UFF':
            self._UFF_args = update_default_kwargs(self._default_UFF_args, kwargs,
                                                 self._to_xyz_core_names[1], self._to_xyz_core_docs[1])
        elif optimizer=='MMFF':
            self._MMFF_args = update_default_kwargs(self._default_MMFF_args, kwargs,
                                                 self._to_xyz_core_names[0], self._to_xyz_core_docs[0])

    def _to_xyz_pybel(self):
        """
        The internal function extracts the xyz coordinates from a pre-built pybel molecule object.
        """
        geometry = np.array([atom.coords for atom in self.pybel_molecule])
        atomic_nums = np.array([atom.atomicnum for atom in self.pybel_molecule])
        # pybel atom types are extended
        Z = {44: 'Ru',
             75: 'Re',
             104: 'Rf',
             111: 'Rg',
             88: 'Ra',
             37: 'Rb',
             86: 'Rn',
             45: 'Rh',
             4: 'Be',
             56: 'Ba',
             107: 'Bh',
             83: 'Bi',
             97: 'Bk',
             35: 'Br',
             1: 'H',
             15: 'P',
             76: 'Os',
             99: 'Es',
             80: 'Hg',
             32: 'Ge',
             64: 'Gd',
             31: 'Ga',
             59: 'Pr',
             78: 'Pt',
             94: 'Pu',
             6: 'C',
             82: 'Pb',
             91: 'Pa',
             46: 'Pd',
             48: 'Cd',
             84: 'Po',
             61: 'Pm',
             108: 'Hs',
             115: 'Uup',
             117: 'Uus',
             118: 'Uuo',
             67: 'Ho',
             72: 'Hf',
             19: 'K',
             2: 'He',
             101: 'Md',
             12: 'Mg',
             42: 'Mo',
             25: 'Mn',
             8: 'O',
             109: 'Mt',
             16: 'S',
             74: 'W',
             30: 'Zn',
             63: 'Eu',
             40: 'Zr',
             68: 'Er',
             28: 'Ni',
             102: 'No',
             11: 'Na',
             41: 'Nb',
             60: 'Nd',
             10: 'Ne',
             93: 'Np',
             87: 'Fr',
             26: 'Fe',
             114: 'Fl',
             100: 'Fm',
             5: 'B',
             9: 'F',
             38: 'Sr',
             7: 'N',
             36: 'Kr',
             14: 'Si',
             50: 'Sn',
             62: 'Sm',
             23: 'V',
             21: 'Sc',
             51: 'Sb',
             106: 'Sg',
             34: 'Se',
             27: 'Co',
             112: 'Cn',
             96: 'Cm',
             17: 'Cl',
             20: 'Ca',
             98: 'Cf',
             58: 'Ce',
             54: 'Xe',
             71: 'Lu',
             55: 'Cs',
             24: 'Cr',
             29: 'Cu',
             57: 'La',
             3: 'Li',
             116: 'Lv',
             81: 'Tl',
             69: 'Tm',
             103: 'Lr',
             90: 'Th',
             22: 'Ti',
             52: 'Te',
             65: 'Tb',
             43: 'Tc',
             73: 'Ta',
             70: 'Yb',
             105: 'Db',
             66: 'Dy',
             110: 'Ds',
             53: 'I',
             92: 'U',
             39: 'Y',
             89: 'Ac',
             47: 'Ag',
             113: 'Uut',
             77: 'Ir',
             95: 'Am',
             13: 'Al',
             33: 'As',
             18: 'Ar',
             79: 'Au',
             85: 'At',
             49: 'In'}
        atomic_symbols = np.array([Z[i] for i in atomic_nums])
        self._xyz = XYZ(geometry, atomic_nums.reshape(-1, 1), atomic_symbols.reshape(-1, 1))

    def visualize(self, filename=None, **kwargs):
        """
        This function visualizes the molecule. If both rdkit and pybel objects are avaialble, the rdkit object
        will be used for visualization.

        Parameters
        ----------
        filename: str, optional (default = None)
            This is the path to the file that you want write the image in it.
            Tkinter and Python Imaging Library are required for writing the image.

        kwargs:
            any extra parameter that you want to pass to the rdkit or pybel draw tool.
            Additional information at:
                - https://www.rdkit.org/docs/source/rdkit.Chem.Draw.html
                - http://openbabel.org/docs/dev/UseTheLibrary/Python_PybelAPI.html#pybel.Molecule.draw

        Returns
        -------
        fig : object
            You will be able to display this object, e.g., inside the Jupyter Notebook.

        """
        engine = self._check_original_molecule()
        if engine == 'rdkit':
            from rdkit.Chem import Draw
            if filename is not None:
                Draw.MolToFile(self.rdkit_molecule, filename, **kwargs)
            else:
                return Draw.MolToImage(self.rdkit_molecule, **kwargs)
        elif engine == 'pybel':
            if filename is not None:
                self.pybel_molecule.draw(show=False, filename=filename, **kwargs)
            else:
                return self.pybel_molecule # it seems that the object alone is displayable




import setuptools
from os import path
import chemml

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

if __name__ == "__main__":
    setuptools.setup(
        name='chemml',
        version=chemml.__version__,
        author='Mojtaba Haghighatlari, Johannes Hachmann',
        author_email='mojtabah@buffalo.edu, hachmann@buffalo.edu',
        # url='https://github.com/hachmannlab/chemml',
        project_urls={
            'Source': 'https://github.com/hachmannlab/chemml',
            'url': 'https://hachmannlab.github.io/chemml/'
        },
        description=
        'A Machine Learning and Informatics Program Suite for the Chemical and Materials Sciences',
        long_description=long_description,
        long_description_content_type="text/markdown",
        keywords=[
            'Machine Learning', 'Data Mining', 'Quantum Chemistry',
            'Materials Science', 'Drug Discovery'
        ],
        license='BSD-3C',
        packages=setuptools.find_packages(),
        include_package_data=True,

        install_requires=[
            'future', 'six',
            'numpy', 'pandas', 'scipy',
            'tensorflow',
            'h5py',
            'scikit-learn',
            'matplotlib>=1.5.1',
            'lxml', 'wget',
            'seaborn',
            'openpyxl', 'ipywidgets'
            ],
        extras_require={
            'docs': [
                'sphinx',
                'sphinxcontrib-napoleon',
                'sphinx_rtd_theme',
                'numpydoc',
                'nbsphinx'
            ],
            'tests': [
                'pytest',
                'pytest-cov',
                'pytest-pep8',
                'tox'
            ],
        },
        tests_require=[
            'pytest',
            'pytest-cov',
            'pytest-pep8',
            'tox',
        ],
        classifiers=[
            'Development Status :: 4 - Beta',
            'Natural Language :: English',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3',
        ],
        zip_safe=False,
    )



import importlib


def get_api(name, library, module):
    """
    This function import a python library and obtain a specific class or function of it by string names.

    Parameters
    ----------
    name: str
        name of the function/class

    library: str
        name of python library

    module: str
        name of the library's module that house the function/class

    Returns
    -------
    tuple
        a tuple of two elements: function/class and their type: 'function'/'class'

    """
    if len(module.strip()) > 0:
        full_module_name = library + "." + module
    else:
        full_module_name = library

    themodule = importlib.import_module(full_module_name)

    api = getattr(themodule, name)

    # must be either class or function
    if isinstance(api, type):
        api_type = 'class'
    else:
        api_type = 'function'

    return (api, api_type)


def get_method(obj, method):
    """
    This function get an instance of a class and obtain a specific method of it by its string name.

    Parameters
    ----------
    obj: instance of a class
        An instance of a class

    method: str
        name of the method

    Returns
    -------
    function
        The method function of the received class.

    """
    return getattr(obj, method)


def get_attributes(obj, attr):
    """
    This function get outputs of a class as a dictionary and substitute values with actual class attributes.

    Parameters
    ----------
    obj: instance of a class
        An instance of a class

    outputs: dict
        A dictionary of class outputs.

    Returns
    -------
    dict:
        A dictionary of class attributes upon request.

    """
    return getattr(obj, attr)


def evaluate_param_value(param_val):
    """
    evaluate the string input value to python data structure

    Parameters
    ----------
    param_val: str
        an entry of type str

    Returns
    -------
    bool
        True if the input can become an integer, False otherwise

    Notes
    -----
    returns 'type' for the entry 'type' since type is a reserved word
    """
    try:
        val = eval(param_val)
        if isinstance(val, type):
            return param_val
        else:
            return val
    except:
        return param_val


def evaluate_inputs(inputs, stack, typ='function'):
    """
    This function get a dictionary of inputs and the stack obj to return the parameters and their values.

    Parameters
    ----------
    inputs: dict
        The dictionary of input parameters to a class or function.

    stack: instance of Stack class
        The only container of the entire send/recv token data in the wrapper.

    Returns
    -------
    dict
        A dictionary of all input parameters with their values.

    Notes
    -----
    Some notes for developers:
        - if var name is `*args`, delete it from input dictionary and return it separately
        - if var name is `obj` in a class, that's an instance of the same class, delete it from input dictionary and
            return it separately
        - if there are more than one input per variable (more than two @ sign), keep them as a list. This is the case for:
            - *args in a function
            - a variable that can receive a list of data arrays
        - if var name is **kwargs, make sure its value is a dictionary, then delete it , and add it to the inputs

    """
    args = None
    obj = None
    for var in inputs:
        val = evaluate_param_value(inputs[var])

        # string values with @
        if isinstance(val, str) and val[0] == '@' and val.count('@') % 2 == 0:
            val_array = receiver(val, stack)
            if var == "*args":
                args = val_array
            elif var == "obj" and typ == "class":
                obj = val_array[0]
            else:
                if len(val_array)==1:
                    val = val_array[0]
                else:
                    val = val_array
        inputs[var] = val

        # kwargs and dictionary
        if var=='**kwargs' and isinstance(val, dict):
            for item in val:
                item_val = evaluate_param_value(val[item])

                # string and receive from input/output
                if isinstance(item_val, str) and item_val[0] == '@' and item_val.count('@') % 2 == 0:
                    val_array = receiver(item_val, stack)
                    if len(val_array) == 1:
                        val = val_array[0]
                    else:
                        val = val_array

                inputs[item] = val

    if args is not None:
        del inputs['*args']

    if obj is not None:
        del inputs['obj']

    return {'kwargs':inputs, 'args':args, 'obj': obj}


def receiver(val, stack):
    temp = val.strip().split('@')[1:]  # "@ID2@df" >> ['', 'ID2', 'df']
    val_array = []
    for item in zip(temp[0::2], temp[1::2]):
        val_array.append(stack.pull(item))
    return val_array


# __name__ = "chemml"
__version__ = "1.0"
__author__ = ["Aditya Sonpal (adityaso@buffalo.edu)", "Garuav Vishwakarma (gvishwak@buffalo.edu) ", "Aatish Pradhan (aatishpr@buffalo.edu)","Mojtaba Haghighatlari (mojtabah@buffalo.edu)", "Johannes Hachmann (hachmann@buffalo.edu)"]


# import sys
# sys.dont_write_bytecode = True



"""
Collection of simple preprocessing wrappers only for the sake of user friendliness.
"""

from .miscellaneous import SplitColumns
from .miscellaneous import SaveCSV
from .miscellaneous import SaveFile
from .miscellaneous import SaveHDF5